from pathlib import Path
from pycivil.EXAUtils.strand7PostPro import PostProcessor

def test_001(tmp_path: Path):
    pp = PostProcessor()

    workPath = str(Path(__file__).resolve().parent)+"/strand7_post_pro/001"
    pp.setWorkingPath(workPath)
    outPath = Path(workPath) / Path("out")
    outPath.mkdir(parents=True, exist_ok=True)

    pp.parse()

    pp.readResults(Path("Elements_Results_RC_SLU.txt"), "WA")
    pp.results_checker(None, check_required='SLU_NM')
    pp.export_checks_to_files(Path("Output_SLU_NM_RC_SLU"), True)
    pp.plot_figure("Elements_Results_RC_SLU.txt", "WA", "SLU_NM")

    pp.readResults(Path("Elements_Results_RC_SLV.txt"), "WA")
    pp.results_checker(1001, check_required='SLU_NM')
    pp.export_checks_to_files(Path("Output_SLU_NM_RC_SLV"), True)
    pp.plot_figure("Elements_Results_RC_SLV.txt", "WA", "SLU_NM")

    pp.readResults(Path("Elements_Results_Plates_Force_SLU.txt"), "FORCE")
    pp.results_checker(1001, check_required='SLU_T')
    pp.export_checks_to_files(Path("Output_Plates_Force_SLU_T_SLU"), True)
    pp.plot_figure("Elements_Results_Plates_Force_SLU.txt", "FORCE", "SLU_T")

    pp.readResults(Path("Elements_Results_Plates_Force_SLV.txt"), "FORCE")
    pp.results_checker(1001, check_required='SLU_T')
    pp.export_checks_to_files(Path("Output_Plates_Force_SLU_T_SLV"), True)
    pp.plot_figure("Elements_Results_Plates_Force_SLV.txt", "FORCE", "SLU_T", 1)

    pp.readResults(Path("Elements_Results_RC_SLE_R.txt"), "WA")
    pp.results_checker(1001, check_required='SLE_NM')
    pp.export_checks_to_files(Path("Output_RC_SLE_NM_SLE_R"), True)
    pp.plot_figure("Elements_Results_RC_SLE_R.txt", "WA", "SLE_NM")

    pp.readResults(Path("Elements_Results_RC_SLE_QP.txt"), "WA")
    pp.results_checker(1001, check_required='SLE_NM')
    pp.export_checks_to_files(Path("Output_RC_SLE_NM_SLE_QP"), True)
    pp.plot_figure("Elements_Results_RC_SLE_QP.txt", "WA", "SLE_NM")

    pp.readResults(Path("Elements_Results_RC_SLE_F.txt"), "WA")
    pp.results_checker(1001, check_required='SLE_F')
    pp.plot_figure("Elements_Results_RC_SLE_F.txt", "WA", "SLE_F")

    pp.readResults(Path("Elements_Results_RC_SLE_QP.txt"), "WA")
    pp.results_checker(1001, check_required='SLE_F')
    pp.plot_figure("Elements_Results_RC_SLE_QP.txt", "WA", "SLE_F")



from typing import Any


class EXAExceptions(Exception):
    def __init__(self, error: str, message: str, number: Any = None) -> None:
        self.error = error
        self.message = message
        self.number = number



import json

import pytest
from pycivil.EXAUtils.EXAExceptions import EXAExceptions
from pycivil.EXAGeometry.geometry import twoPointsDivide, twoPointsOffset
from pycivil.EXAStructural.loads import ForcesOnSection
from pycivil.EXAStructural.rcgensolver.templateRCGen import *
from pycivil.EXAUtils.report import (
    ReportDriverEnum,
    Reporter,
    ReportTemplateEnum,
    getTemplatesPath
)
from pycivil.EXAStructural.codes import Code

def ppr(jsonFrag):
    print(json.dumps(jsonFrag.model_dump(mode="json"), indent=4))


def test_001_base(tmp_path: Path):
    # Build a modeler
    modeler = RCGenSectionsModeler()
    modeler.setJobPath(os.path.dirname(__file__))

    idSectionModel = modeler.addSectionModel(key="section_1", current=True)
    assert idSectionModel != -1
    assert modeler.addSectionModel(key="section_1", current=True) == -1

    # Build solid part
    w = 300
    h = 600
    shapeRectangular = ShapeRect(width=w, height=h)
    modeler.addSolid(shape=shapeRectangular)

    # Add single rebars with point
    cc = 50
    rebarsDiameterTOP = 8
    rebarsDiameterBOT = 24
    barBL = ShapeCircle(
        radius=rebarsDiameterBOT / 2,
        center=shapeRectangular.getShapePoint("BL") + Point2d(cc, cc),
    )
    modeler.addRebar(shape=barBL)
    barBR = ShapeCircle(
        radius=rebarsDiameterBOT / 2,
        center=shapeRectangular.getShapePoint("BR") + Point2d(-cc, cc),
    )
    modeler.addRebar(shape=barBR)
    barTR = ShapeCircle(
        radius=rebarsDiameterTOP / 2,
        center=shapeRectangular.getShapePoint("TR") + Point2d(-cc, -cc),
    )
    modeler.addRebar(shape=barTR)
    barTL = ShapeCircle(
        radius=rebarsDiameterTOP / 2,
        center=shapeRectangular.getShapePoint("TL") + Point2d(cc, -cc),
    )
    modeler.addRebar(shape=barTL)

    # Add multiple rebar with point
    barsBOT_1, barsBOT_2 = twoPointsOffset(barBL.center(), barBR.center())
    barsBOT = twoPointsDivide(barsBOT_1, barsBOT_2, 4, False)
    barsTOP_1, barsTOP_2 = twoPointsOffset(barTR.center(), barTL.center())
    barsTOP = twoPointsDivide(barsTOP_1, barsTOP_2, 4, False)
    modeler.addRebarsGroup(shape=barBL, points=barsBOT)
    modeler.addRebarsGroup(shape=barTL, points=barsTOP)

    # Setting code
    code = Code("NTC2018")

    # Build concrete material by law
    concreteMat = Concrete(descr="Concrete as NTC2018 Law")
    concreteMat.setByCode(code, "C30/37")

    # Build rebar material by law
    rebarMat = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat.setByCode(code, "B450C")

    # Assign materials
    idMatCon = modeler.addConcreteLaw(mat=concreteMat)
    modeler.assignConcreteLawToCurrentModel(idMatCon)
    idMatReb = modeler.addRebarLaw(mat=rebarMat)
    modeler.assignRebarLawToCurrentModel(idMatReb)

    rebarArea = modeler.calcRebarArea()
    concreteArea = modeler.calcConcreteArea()

    fcd = concreteMat.get_fck() * concreteMat.get_alphacc() / concreteMat.get_gammac()
    fyd = rebarMat.get_fsy() / rebarMat.get_gammas()

    N_Rd_comp = -(fyd * rebarArea + fcd * concreteArea)
    print(f"N_Rd_comp (analytical)= {N_Rd_comp}")
    N_Rd_tens = fyd * rebarArea
    print(f"N_Rd_tens (analytical)= {N_Rd_tens}")

    assert modeler.run(opt=Analysis.BOUNDING_SLU)
    results = modeler.getModelOutput().sectionResults[idSectionModel]

    # Test for minimum bounding relative to N force
    x0 = results.boundingSLU.minForces.Fz
    x1 = N_Rd_comp
    print(f"N_Rd_comp (from domain)= {x0}")
    assert x1 == pytest.approx(x0, 1e-12)

    # Test for maximum bounding relative to N force
    x0 = results.boundingSLU.maxForces.Fz
    x1 = N_Rd_tens
    print(f"N_Rd_tens (from domain)= {x0}")
    assert x1 == pytest.approx(x0, 1e-12)

    # Moments for later use
    Mxmin = results.boundingSLU.minForces.Mx
    print(f"Mxmin (from domain)= {Mxmin}")
    Mxmax = results.boundingSLU.maxForces.Mx
    print(f"Mxmax (from domain)= {Mxmax}")
    Mymin = results.boundingSLU.minForces.My
    print(f"Mymin (from domain)= {Mymin}")
    Mymax = results.boundingSLU.maxForces.My
    print(f"Mymax (from domain)= {Mymax}")

    nbForcesGenerated = 10
    forces = {}
    for i in range(nbForcesGenerated):
        Fzi = (N_Rd_tens - N_Rd_comp) / (nbForcesGenerated + 2) * (i + 1) + N_Rd_comp
        forces[i + 1] = ForcesOnSection(id=i + 1, Fx=0, Fy=0, Fz=Fzi, Mx=0, My=0, Mz=0)

    nbForcesNull = 4
    for i in range(nbForcesNull):
        forces[len(forces) + 100] = ForcesOnSection(
            id=len(forces) + 100,
            Fx=0,
            Fy=0,
            Fz=0,
            Mx=0,
            My=0,
            Mz=0,
            limitState=LimitState.ULTIMATE,
        )

    # Model without forces make error
    assert not modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)
    results = modeler.getModelOutput().sectionResults[idSectionModel]
    assert "|ERR --->| Forces with lenght=0 !!! Stop." == results.domainSLU.logs[0]

    # Model with forces go well
    modeler.assignForces(forces)
    assert modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)
    results = modeler.getModelOutput().sectionResults[idSectionModel]

    outfile = Path(__file__).parent.joinpath("test_001_base_nr_1.json")
    jsonObject = json.loads(outfile.read_text())
    assert results == Results(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile.write_text(json.dumps(results.model_dump(), indent=4))

    assert "|INF| Successfully assigned forces for analysis ...", results.domainSLU.logs[
        0
    ]
    assert (
        f"|WRN ->| All {len(forces)} forces are null or not SLU !!!"
    ), results.domainSLU.logs[1]
    assert (
        f"|INF| Nb. {nbForcesNull}/{len(forces)} forces are null !!!"
    ), results.domainSLU.logs[2]
    assert (
        f"|INF| Nb. {len(forces)}/{len(forces)} forces are not ULS !!!"
    ), results.domainSLU.logs[3]

    # Assign ultimate limit state
    for _k, f in forces.items():
        f.limitState = LimitState.ULTIMATE

    assert modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)
    results = modeler.getModelOutput().sectionResults[idSectionModel]

    outfile = Path(__file__).parent.joinpath("test_001_base_nr_2.json")
    jsonObject = json.loads(outfile.read_text())
    assert results == Results(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile.write_text(json.dumps(results.model_dump(), indent=4))

    for v in results.domainSLU.loadsCheckLogs.values():
        assert "Domain ok", v.logs[0]
        assert "N is contained in bounding", v.logs[1]
        assert "Moments Mx and My are nulls. Banal solution", v.logs[2]

    i = 0
    for force in forces.values():
        force.Mx = (Mxmax - Mxmin) / (nbForcesGenerated + 2) * (i + 1) + Mxmin
        force.My = (Mymax - Mymin) / (nbForcesGenerated + 2) * (i + 1) + Mymin
        i += 1
    assert modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)

    modeler.plot(pfx="test_001_base_compact_")
    modeler.plot(pfx="test_001_base_extended_", onlyWorst=False)

    # results = modeler.getModelOutput().sectionResults[idSectionModel]
    results = modeler.getModelOutput()
    outfile = Path(__file__).parent.joinpath("test_001_base_nr_3.json")
    jsonObject = json.loads(outfile.read_text())
    assert results == RCGenSectionsOutput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile.write_text(results.model_dump_json(indent=4))

    input_model = modeler.exportModelInput()
    input_file = Path(__file__).parent.joinpath("test_001_base_input_nr_3.json")
    jsonObject = json.loads(input_file.read_text())
    assert input_model == RCGenSectionsInput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # input_file.write_text(input_model.model_dump_json(indent=4))


def test_002_domain_multisections(tmp_path: Path):
    modeler = RCGenSectionsModeler()
    with pytest.raises(EXAExceptions):
        modeler.run(opt=Analysis.BOUNDING_SLU)

    modeler.setJobPath(os.path.dirname(__file__))

    # ----------
    # SECTION #1
    # ----------
    idSec1 = modeler.addSectionModel(key="section_1", current=True)
    modeler.addSolid(shape=ShapeRect(width=300, height=600))

    # Geometry
    cc1 = 50
    nb1 = 4
    dTOP1 = 8
    dBOT1 = 24
    lpBOT1 = modeler.getShapeVertex("BL") + Point2d(cc1, cc1)
    rpBOT1 = modeler.getShapeVertex("BR") + Point2d(-cc1, cc1)
    lpTOP1 = modeler.getShapeVertex("TL") + Point2d(cc1, -cc1)
    rpTOP1 = modeler.getShapeVertex("TR") + Point2d(-cc1, -cc1)
    pointsBOT1 = twoPointsDivide(lpBOT1, rpBOT1, nb1, True)
    pointsTOP1 = twoPointsDivide(lpTOP1, rpTOP1, nb1, True)
    modeler.addRebarsGroup(diameter=dBOT1, points=pointsBOT1)
    modeler.addRebarsGroup(diameter=dTOP1, points=pointsTOP1)

    # Setting code
    code1 = Code("NTC2008")

    # Build concrete material by law
    concreteMat1 = Concrete(descr="Concrete as NTC2008 Law")
    concreteMat1.setByCode(code1, "C32/40")

    # Build rebar material by law
    rebarMat1 = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat1.setByCode(code1, "B450C")

    # Assign materials
    idMatConcrete1 = modeler.addConcreteLaw(mat=concreteMat1)
    idMatRebar1 = modeler.addRebarLaw(mat=rebarMat1)
    modeler.assignConcreteLawToCurrentModel(idm=idMatConcrete1)
    modeler.assignRebarLawToCurrentModel(idm=idMatRebar1)

    assert modeler.run(opt=Analysis.BOUNDING_SLU)
    results = modeler.getModelOutput().sectionResults[idSec1]

    # Moments for generator
    Fzmin = results.boundingSLU.minForces.Fz
    Fzmax = results.boundingSLU.maxForces.Fz
    Mxmin = results.boundingSLU.minForces.Mx
    Mxmax = results.boundingSLU.maxForces.Mx
    Mymin = results.boundingSLU.minForces.My
    Mymax = results.boundingSLU.maxForces.My

    nbForcesGenerated = 10
    forces = {}
    for i in range(nbForcesGenerated):
        Fzi = (Fzmax - Fzmin) / (nbForcesGenerated + 2) * (i + 1) + Fzmin
        Mxi = (Mxmax - Mxmin) / (nbForcesGenerated + 2) * (i + 1) + Mxmin
        Myi = (Mymax - Mymin) / (nbForcesGenerated + 2) * (i + 1) + Mymin
        forces[i + 1] = ForcesOnSection(
            id=i + 1, Fz=Fzi, Mx=Mxi, My=Myi, limitState=LimitState.ULTIMATE
        )

    modeler.assignForces(forces)

    # ----------
    # SECTION #2
    # ----------
    idSec2 = modeler.addSectionModel(key="section_2", current=True)
    modeler.addSolid(shape=ShapeRect(width=800, height=1200))

    cc2 = 70
    nb2 = 10
    dTOP2 = 26
    dBOT2 = 26
    lpBOT2 = modeler.getShapeVertex("BL") + Point2d(cc2, cc2)
    rpBOT2 = modeler.getShapeVertex("BR") + Point2d(-cc2, cc2)
    lpTOP2 = modeler.getShapeVertex("TL") + Point2d(cc2, -cc2)
    rpTOP2 = modeler.getShapeVertex("TR") + Point2d(-cc2, -cc2)
    pointsBOT2 = twoPointsDivide(lpBOT2, rpBOT2, nb2, True)
    pointsTOP2 = twoPointsDivide(lpTOP2, rpTOP2, nb2, True)
    modeler.addRebarsGroup(diameter=dBOT2, points=pointsBOT2)
    modeler.addRebarsGroup(diameter=dTOP2, points=pointsTOP2)

    # Setting code
    code2 = Code("NTC2018")

    # Build concrete material by law
    concreteMat2 = Concrete(descr="Concrete as NTC2018 Law")
    concreteMat2.setByCode(code2, "C35/45")

    # Build rebar material by law
    rebarMat2 = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat2.setByCode(code2, "B450C")

    # Assign materials
    idMatConcrete2 = modeler.addConcreteLaw(mat=concreteMat2)
    idMatRebar2 = modeler.addRebarLaw(mat=rebarMat2)
    modeler.assignConcreteLawToCurrentModel(idm=idMatConcrete2)
    modeler.assignRebarLawToCurrentModel(idm=idMatRebar2)

    assert modeler.run(opt=Analysis.BOUNDING_SLU)
    results = modeler.getModelOutput().sectionResults[idSec2]

    # Moments for generator
    Fzmin = results.boundingSLU.minForces.Fz
    Fzmax = results.boundingSLU.maxForces.Fz
    Mxmin = results.boundingSLU.minForces.Mx
    Mxmax = results.boundingSLU.maxForces.Mx
    Mymin = results.boundingSLU.minForces.My
    Mymax = results.boundingSLU.maxForces.My

    nbForcesGenerated = 10
    forces = {}
    for i in range(nbForcesGenerated):
        Fzi = (Fzmax - Fzmin) / (nbForcesGenerated + 2) * (i + 1) + Fzmin
        Mxi = (Mxmax - Mxmin) / (nbForcesGenerated + 2) * (i + 1) + Mxmin
        Myi = (Mymax - Mymin) / (nbForcesGenerated + 2) * (i + 1) + Mymin
        forces[i + 1] = ForcesOnSection(
            id=i + 1, Fz=Fzi, Mx=Mxi, My=Myi, limitState=LimitState.ULTIMATE
        )

    modeler.assignForces(forces)
    modeler.setCurrentModel("section_1")
    assert modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)
    modeler.setCurrentModel("section_2")
    assert modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)

    modeler.logLevel = 3
    modeler.plot(pfx="test_002_multisections_compact_")
    modeler.plot(pfx="test_002_multisections_extended_", onlyWorst=False)

    results = modeler.getModelOutput()
    outfile = Path(__file__).parent.joinpath("test_002_domain_multisections_nr.json")
    jsonObject = json.loads(outfile.read_text())
    assert results == RCGenSectionsOutput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile.write_text(results.model_dump_json(indent=4))

    input_model = modeler.exportModelInput()
    input_file = Path(__file__).parent.joinpath(
        "test_002_domain_multisection_input_nr.json"
    )
    jsonObject = json.loads(input_file.read_text())
    assert input_model == RCGenSectionsInput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # input_file.write_text(input_model.model_dump_json(indent=4))


def test_003_elastic_multisections(tmp_path: Path):
    modeler = RCGenSectionsModeler()
    modeler.logLevel = 3
    modeler.setJobPath(os.path.dirname(__file__))

    # ---------------------------
    # Assign material to sections
    # ---------------------------
    # Setting code
    code = Code("NTC2018")
    # Build concrete material by law
    concreteMat = Concrete(descr="Concrete as NTC2018 Law")
    concreteMat.setByCode(code, "C30/37")
    # Build rebar material by law
    rebarMat = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat.setByCode(code, "B450C")
    # Assign materials
    idMatCon = modeler.addConcreteLaw(mat=concreteMat)
    idMatReb = modeler.addRebarLaw(mat=rebarMat)

    # ------------------------------------
    # SEZIONE [1] exagon_test_numerici.pdf
    # ------------------------------------
    idSec1 = modeler.addSectionModel(key="section_1", current=True)
    modeler.assignConcreteLawToCurrentModel(idMatCon)
    modeler.assignRebarLawToCurrentModel(idMatReb)

    modeler.addSolid(shape=ShapeRect(width=300, height=600))
    vBL = modeler.getShapeVertex("BL")
    vBR = modeler.getShapeVertex("BR")
    vTL = modeler.getShapeVertex("TL")
    vTR = modeler.getShapeVertex("TR")
    modeler.addRebar(center=vBL + Point2d(20, 20), diameter=20)
    modeler.addRebar(center=vBL + Point2d(100, 20), diameter=20)
    modeler.addRebar(center=vBR + Point2d(-20, 20), diameter=20)
    modeler.addRebar(center=vBR + Point2d(-100, 20), diameter=20)
    modeler.addRebar(center=vTL + Point2d(20, -20), diameter=20)
    modeler.addRebar(center=vTL + Point2d(100, -20), diameter=20)
    modeler.addRebar(center=vTR + Point2d(-20, -20), diameter=20)
    modeler.addRebar(center=vTR + Point2d(-100, -20), diameter=20)

    KN = 1000
    KNm = 1000 * 1000
    forces = dict()
    forces[6] = ForcesOnSection(
        id=6,
        Fz=0.0 * KN,
        Mx=1000 * KNm,
        My=1000 * KNm,
        limitState=LimitState.SERVICEABILITY,
    )
    forces[7] = ForcesOnSection(
        id=7,
        Fz=0.0 * KN,
        Mx=100 * KNm,
        My=1000 * KNm,
        limitState=LimitState.SERVICEABILITY,
    )
    forces[8] = ForcesOnSection(
        id=8,
        Fz=0.0 * KN,
        Mx=500 * KNm,
        My=1000 * KNm,
        limitState=LimitState.SERVICEABILITY,
    )
    forces[9] = ForcesOnSection(
        id=9,
        Fz=0.0 * KN,
        Mx=-1000 * KNm,
        My=0.0 * KNm,
        limitState=LimitState.SERVICEABILITY,
    )
    forces[10] = ForcesOnSection(id=10, Fz=0.0 * KN, Mx=500 * KNm, My=1000 * KNm)
    forces[11] = ForcesOnSection(id=11, Fz=0.0 * KN, Mx=-1000 * KNm, My=0.0 * KNm)

    # Without forces runner gives False
    assert not modeler.run(opt=Analysis.ELASTIC_SOLVER)

    # Without run ELASTIC_SOLVER before gives False
    assert not modeler.run(opt=Analysis.CHECK_ELASTIC)

    modeler.assignForces(forces)
    assert modeler.run(opt=Analysis.ELASTIC_SOLVER)

    # Without frequency for force ELASTIC_SOLVER before gives False
    assert not modeler.run(opt=Analysis.CHECK_ELASTIC)

    forces[6].frequency = Frequency.CHARACTERISTIC
    forces[7].frequency = Frequency.CHARACTERISTIC
    forces[8].frequency = Frequency.CHARACTERISTIC
    forces[9].frequency = Frequency.QUASI_PERMANENT

    modeler.assignForces(forces)
    assert modeler.run(opt=Analysis.CHECK_ELASTIC)

    modeler.plot(pfx="test_003_elastic_multisections_")
    modeler.plot(pfx="test_003_elastic_multisections_extended_", onlyWorst=False)

    outModel = modeler.getModelOutput()
    assert type(outModel) == RCGenSectionsOutput

    outModelCasted = RCGenSectionsOutput.model_validate(outModel.model_dump())
    resultsForForce = outModelCasted.sectionResults[
        idSec1
    ].elasticSolver.elasticResults[6]
    assert pytest.approx(resultsForForce.sigmacMin, rel=1e-4) == -199.6
    assert pytest.approx(resultsForForce.sigmacMax, rel=1e-4) == 0.0
    assert pytest.approx(resultsForForce.sigmasMin, rel=1e-4) == -2464
    assert pytest.approx(resultsForForce.sigmasMax, rel=1e-4) == 5426

    resultsForForce = outModelCasted.sectionResults[
        idSec1
    ].elasticSolver.elasticResults[7]
    assert pytest.approx(resultsForForce.sigmacMin, rel=1e-3) == -62.2
    assert pytest.approx(resultsForForce.sigmacMax, rel=1e-3) == 0.0
    assert pytest.approx(resultsForForce.sigmasMin, rel=1e-3) == -819
    assert pytest.approx(resultsForForce.sigmasMax, rel=1e-3) == 1755

    resultsForForce = outModelCasted.sectionResults[
        idSec1
    ].elasticSolver.elasticResults[8]
    assert pytest.approx(resultsForForce.sigmacMin, rel=1e-4) == -128.9
    assert pytest.approx(resultsForForce.sigmacMax, rel=1e-4) == 0.0
    assert pytest.approx(resultsForForce.sigmasMin, rel=1e-3) == -1636
    assert pytest.approx(resultsForForce.sigmasMax, rel=1e-4) == 3263

    resultsForForce = outModelCasted.sectionResults[
        idSec1
    ].elasticSolver.elasticResults[9]
    assert pytest.approx(resultsForForce.sigmacMin, rel=1e-3) == -136.3
    assert pytest.approx(resultsForForce.sigmacMax, rel=1e-4) == 0.0
    assert pytest.approx(resultsForForce.sigmasMin, rel=1e-3) == -1580
    assert pytest.approx(resultsForForce.sigmasMax, rel=1e-3) == 4448

    outfile = Path(__file__).parent.joinpath("test_003_elastic_multisections_nr.json")
    jsonObject = json.loads(outfile.read_text())
    assert outModel == RCGenSectionsOutput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # outfile.write_text(outModelCasted.model_dump_json(indent=4))

    input_model = modeler.exportModelInput()
    input_file = Path(__file__).parent.joinpath(
        "test_003_elastic_multisections_input_nr.json"
    )
    jsonObject = json.loads(input_file.read_text())
    assert input_model == RCGenSectionsInput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # input_file.write_text(input_model.model_dump_json(indent=4))


# Test 2nd way to shape a section with general commands.
# Section is the same of test_001.
#
def test_004_base_general(tmp_path: Path):
    # Build a modeler
    modeler = RCGenSectionsModeler()
    modeler.setJobPath(os.path.dirname(__file__))

    idSectionModel = modeler.addSectionModel(key="section_1", current=True, ids=1)
    assert idSectionModel == 1
    # Test same key
    assert modeler.addSectionModel(key="section_1", current=True) == -1
    # Test same id
    assert modeler.addSectionModel(key="section_2", current=True, ids=1) == -1

    nodeBL = Node2d(x=-150, y=-300, idn=1)
    nodeBR = Node2d(x=+150, y=-300, idn=2)
    nodeTR = Node2d(x=+150, y=+300, idn=3)
    nodeTL = Node2d(x=-150, y=+300, idn=4)
    modeler.addNode(nodeBL)
    modeler.addNode(nodeBR)
    modeler.addNode(nodeTR)
    modeler.addNode(nodeTL)

    modeler.addTriangle(nodeBL.idn, nodeBR.idn, nodeTL.idn, 1)
    modeler.addTriangle(nodeBR.idn, nodeTR.idn, nodeTL.idn, 2)

    # Add single rebars with point
    cc = 50
    rebarsDiameterTOP = 8
    rebarsDiameterBOT = 24
    barBL = ShapeCircle(radius=rebarsDiameterBOT / 2, center=nodeBL + Point2d(cc, cc))
    modeler.addRebar(shape=barBL, idn=5, idr=1)
    barBR = ShapeCircle(radius=rebarsDiameterBOT / 2, center=nodeBR + Point2d(-cc, cc))
    modeler.addRebar(shape=barBR, idn=6, idr=2)
    barTR = ShapeCircle(radius=rebarsDiameterTOP / 2, center=nodeTR + Point2d(-cc, -cc))
    modeler.addRebar(shape=barTR, idn=7, idr=3)
    barTL = ShapeCircle(radius=rebarsDiameterTOP / 2, center=nodeTL + Point2d(cc, -cc))
    modeler.addRebar(shape=barTL, idn=8, idr=4)

    # Add multiple rebar with point
    barsBOT_1, barsBOT_2 = twoPointsOffset(barBL.center(), barBR.center())
    barsBOT = twoPointsDivide(barsBOT_1, barsBOT_2, 4, False)
    barsTOP_1, barsTOP_2 = twoPointsOffset(barTR.center(), barTL.center())
    barsTOP = twoPointsDivide(barsTOP_1, barsTOP_2, 4, False)
    modeler.addRebarsGroup(
        shape=barBL, points=barsBOT, id_nodes=[9, 10, 11], id_rebars=[7, 6, 5]
    )
    modeler.addRebarsGroup(
        shape=barTL, points=barsTOP, id_nodes=[12, 13, 14], id_rebars=[8, 9, 10]
    )

    modeler.plot(pfx="test_004_base_compact_")

    # Setting code
    code = Code("NTC2018")

    # Build concrete material by law
    concreteMat = Concrete(descr="Concrete as NTC2018 Law")
    concreteMat.setByCode(code, "C30/37")

    # Build rebar material by law
    rebarMat = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat.setByCode(code, "B450C")

    # Assign materials
    idMatCon = modeler.addConcreteLaw(mat=concreteMat, idm=1)
    assert idMatCon == 1
    modeler.assignConcreteLawToCurrentModel(idMatCon)
    idMatReb = modeler.addRebarLaw(mat=rebarMat, idm=2)
    assert idMatReb > 0
    modeler.assignRebarLawToCurrentModel(idMatReb)

    assert modeler.run(opt=Analysis.BOUNDING_SLU)
    results = modeler.getModelOutput().sectionResults[idSectionModel]

    N_Rd_comp = results.boundingSLU.minForces.Fz
    N_Rd_tens = results.boundingSLU.maxForces.Fz

    assert N_Rd_comp == pytest.approx(-4043455.0915585444, 1e-12)
    assert N_Rd_tens == pytest.approx(983455.0915585441, 1e-12)

    # Moments for later use
    Mxmin = results.boundingSLU.minForces.Mx
    Mxmin_expected = -468694010.4534114
    Mxmax_expected = 468694010.45341134
    Mymin_expected = -160832599.8557186
    Mymax_espected = 160832599.8557186
    assert Mxmin == pytest.approx(Mxmin_expected, 1e-12)
    Mxmax = results.boundingSLU.maxForces.Mx
    assert Mxmax == pytest.approx(Mxmax_expected, 1e-12)
    Mymin = results.boundingSLU.minForces.My
    assert Mymin == pytest.approx(Mymin_expected, 1e-12)
    Mymax = results.boundingSLU.maxForces.My
    assert Mymax == pytest.approx(Mymax_espected, 1e-12)

    input_model = modeler.exportModelInput()
    input_file = Path(__file__).parent.joinpath("test_004_base_general_input_nr.json")
    jsonObject = json.loads(input_file.read_text())
    assert input_model == RCGenSectionsInput(**jsonObject)
    # ------------------------------------------------
    # using this for generate not regression json file
    # ------------------------------------------------
    # input_file.write_text(input_model.model_dump_json(indent=4))

#
# This test take tests above and rebuilds models
#
def test_005_solver_from_model(tmp_path: Path):
    def in_and_out(file_name: str):
        #
        # Read json input file
        #
        input_file = Path(__file__).parent.joinpath(file_name)
        jsonObject = json.loads(input_file.read_text())
        input_model = RCGenSectionsInput(**jsonObject)
        #
        # Build modeler with minimal settings
        #
        modeler = RCGenSectionsModeler()
        modeler.logLevel = 3
        assert modeler.setJobPath(os.path.dirname(__file__))
        assert modeler.setModelInput(input_model)
        #
        # Export and test
        #
        input_model = modeler.exportModelInput()
        input_file = Path(__file__).parent.joinpath(file_name)
        jsonObject = json.loads(input_file.read_text())
        assert input_model == RCGenSectionsInput(**jsonObject)

    ###########################################################################
    # File:                   test_001_base_input_nr_3.json
    ###########################################################################
    in_and_out("test_001_base_input_nr_3.json")
    ###########################################################################
    # File:            test_002_domain_multisection_input_nr.json
    ###########################################################################
    in_and_out("test_002_domain_multisection_input_nr.json")
    ###########################################################################
    # File:            test_003_elastic_multisections_input_nr.json
    ###########################################################################
    in_and_out("test_003_elastic_multisections_input_nr.json")
    ###########################################################################
    # File:                test_004_base_general_input_nr.json
    ###########################################################################
    in_and_out("test_004_base_general_input_nr.json")

def input_model_from_file(file_name: str) -> RCGenSectionsInput:
    #
    # Read json input file
    #
    input_file = Path(__file__).parent.joinpath(file_name)
    jsonObject = json.loads(input_file.read_text())
    return RCGenSectionsInput(**jsonObject)

def output_model_from_file(file_name: str) -> RCGenSectionsOutput:
    #
    # Read json output file
    #
    input_file = Path(__file__).parent.joinpath(file_name)
    jsonObject = json.loads(input_file.read_text())
    return RCGenSectionsOutput(**jsonObject)

def test_006_report_builder(tmp_path: Path):
    i_model = input_model_from_file("test_001_base_input_nr_3.json")
    o_model = output_model_from_file("test_001_base_nr_3.json")
    solver_name = "./RCGen"
    job_path =  Path(os.path.dirname(__file__)) / Path(solver_name)
    rb = ReportBuilder(i_model, o_model, job_path)
    reporter = Reporter(getTemplatesPath())
    reporter.linkFragments(
        driver=ReportDriverEnum.PDFLATEX,
        template=ReportTemplateEnum.TEX_ENG_CAL,
        builder=rb)
    reporter.compileDocument(path=str(tmp_path), fileName="test_006_1_report_builder")

    i_model = input_model_from_file("test_002_domain_multisection_input_nr.json")
    o_model = output_model_from_file("test_002_domain_multisections_nr.json")
    solver_name = "./RCGen"
    job_path =  Path(os.path.dirname(__file__)) / Path(solver_name)
    rb = ReportBuilder(i_model, o_model, job_path)
    reporter = Reporter(getTemplatesPath())
    reporter.linkFragments(
        driver=ReportDriverEnum.PDFLATEX,
        template=ReportTemplateEnum.TEX_ENG_CAL,
        builder=rb)
    reporter.compileDocument(path=str(tmp_path), fileName="test_006_2_report_builder")

    i_model = input_model_from_file("test_003_elastic_multisections_input_nr.json")
    o_model = output_model_from_file("test_003_elastic_multisections_nr.json")
    solver_name = "./RCGen"
    job_path =  Path(os.path.dirname(__file__)) / Path(solver_name)
    rb = ReportBuilder(i_model, o_model, job_path)
    reporter = Reporter(getTemplatesPath())
    reporter.linkFragments(
        template=ReportTemplateEnum.TEX_ENG_CAL,
        builder=rb)
    reporter.compileDocument(path=str(tmp_path), fileName="test_006_3_report_builder")

def test_007_multi_section_multi_check(tmp_path: Path):
    modeler = RCGenSectionsModeler()
    # modeler.logLevel = 3
    modeler.setJobPath(tmp_path)


    modeler.addSectionModel(key="section_1", current=True)

    nodeBL_1 = Node2d(x=-150, y=-300)
    nodeBR_1 = Node2d(x=+150, y=-300)
    nodeTR_1 = Node2d(x=+150, y=+300)
    nodeTL_1 = Node2d(x=-150, y=+300)

    id_nodeBL_1 = modeler.addNode(nodeBL_1)
    id_nodeBR_1 = modeler.addNode(nodeBR_1)
    id_nodeTR_1 = modeler.addNode(nodeTR_1)
    id_nodeTL_1 = modeler.addNode(nodeTL_1)

    modeler.addTriangle(id_nodeBL_1, id_nodeBR_1, id_nodeTL_1)
    modeler.addTriangle(id_nodeBR_1, id_nodeTR_1, id_nodeTL_1)

    cc = 50
    rebarsDiameterTOP = 8
    rebarsDiameterBOT = 24

    rebar_center_BL_1 = nodeBL_1 + Point2d(cc, cc)
    rebar_center_BR_1 = nodeBR_1 + Point2d(-cc, cc)
    rebar_center_TR_1 = nodeTR_1 + Point2d(-cc, -cc)
    rebar_center_TL_1 = nodeTL_1 + Point2d(cc, -cc)
    modeler.addRebar(center=rebar_center_BL_1, diameter=rebarsDiameterBOT)
    modeler.addRebar(center=rebar_center_BR_1, diameter=rebarsDiameterBOT)
    modeler.addRebar(center=rebar_center_TR_1, diameter=rebarsDiameterTOP)
    modeler.addRebar(center=rebar_center_TL_1, diameter=rebarsDiameterTOP)

    modeler.addSectionModel(key="section_2", current=True)

    nodeBL_2 = Node2d(x=-150, y=-300)
    nodeBR_2 = Node2d(x=+150, y=-300)
    nodeML_2 = Node2d(x=+150, y=+100)
    nodeMR_2 = Node2d(x=+300, y=+100)
    nodeTR_2 = Node2d(x=+300, y=+300)
    nodeTL_2 = Node2d(x=-150, y=+300)

    id_nodeBL_2 = modeler.addNode(nodeBL_2)
    id_nodeBR_2 = modeler.addNode(nodeBR_2)
    id_nodeML_2 = modeler.addNode(nodeML_2)
    id_nodeMR_2 = modeler.addNode(nodeMR_2)
    id_nodeTR_2 = modeler.addNode(nodeTR_2)
    id_nodeTL_2 = modeler.addNode(nodeTL_2)

    modeler.addTriangle(id_nodeBL_2, id_nodeML_2, id_nodeTL_2)
    modeler.addTriangle(id_nodeBL_2, id_nodeBR_2, id_nodeML_2)
    modeler.addTriangle(id_nodeTL_2, id_nodeML_2, id_nodeTR_2)
    modeler.addTriangle(id_nodeML_2, id_nodeMR_2, id_nodeTR_2)

    rebar_center_BL_2 = nodeBL_2 + Point2d(cc, cc)
    rebar_center_BR_2 = nodeBR_2 + Point2d(-cc, cc)
    rebar_center_TR_2 = nodeTR_2 + Point2d(-cc, -cc)
    rebar_center_TL_2 = nodeTL_2 + Point2d(cc, -cc)
    modeler.addRebar(center=rebar_center_BL_2, diameter=rebarsDiameterBOT)
    modeler.addRebar(center=rebar_center_BR_2, diameter=rebarsDiameterBOT)
    modeler.addRebar(center=rebar_center_TR_2, diameter=rebarsDiameterTOP)
    modeler.addRebar(center=rebar_center_TL_2, diameter=rebarsDiameterTOP)

    # Add multiple rebar with point
    modeler.setCurrentModel("section_1")
    barsBOT_1_1, barsBOT_2_1 = twoPointsOffset(rebar_center_BL_1, rebar_center_BR_1)
    barsBOT_1 = twoPointsDivide(barsBOT_1_1, barsBOT_2_1, 4, False)
    barsTOP_1_1, barsTOP_2_1 = twoPointsOffset(rebar_center_TR_1, rebar_center_TL_1)
    barsTOP_1 = twoPointsDivide(barsTOP_1_1, barsTOP_2_1, 4, False)
    modeler.addRebarsGroup(
        diameter=rebarsDiameterBOT, points=barsBOT_1
    )
    modeler.addRebarsGroup(
        diameter=rebarsDiameterTOP, points=barsTOP_1
    )

    # Add multiple rebar with point
    modeler.setCurrentModel("section_2")
    barsBOT_1_2, barsBOT_2_2 = twoPointsOffset(rebar_center_BL_2, rebar_center_BR_2)
    barsBOT_2 = twoPointsDivide(barsBOT_1_2, barsBOT_2_2, 4, False)
    barsTOP_1_2, barsTOP_2_2 = twoPointsOffset(rebar_center_TR_2, rebar_center_TL_2)
    barsTOP_2 = twoPointsDivide(barsTOP_1_2, barsTOP_2_2, 4, False)
    modeler.addRebarsGroup(
        diameter=rebarsDiameterBOT, points=barsBOT_2
    )
    modeler.addRebarsGroup(
        diameter=rebarsDiameterTOP, points=barsTOP_2
    )

    # Setting code
    code = Code("NTC2018")

    # Build concrete material by law
    concreteMat = Concrete(descr="Concrete as NTC2018 Law")
    concreteMat.setByCode(code, "C30/37")

    # Build rebar material by law
    rebarMat = ConcreteSteel(descr="Steel as NTC2018 Law")
    rebarMat.setByCode(code, "B450C")

    # Assign materials
    idMatCon = modeler.addConcreteLaw(mat=concreteMat, idm=1)
    idMatReb = modeler.addRebarLaw(mat=rebarMat, idm=2)

    modeler.setCurrentModel("section_1")
    modeler.assignConcreteLawToCurrentModel(idMatCon)
    modeler.assignRebarLawToCurrentModel(idMatReb)
    modeler.setCurrentModel("section_2")
    modeler.assignConcreteLawToCurrentModel(idMatCon)
    modeler.assignRebarLawToCurrentModel(idMatReb)

    KN = 1000
    KNm = 1000 * 1000
    forces = dict()
    forces[1] = ForcesOnSection(
        id=1, Fz=-100.0 * KN, Mx=70 * KNm, My=70 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.CHARACTERISTIC
    )
    forces[2] = ForcesOnSection(
        id=2, Fz=-50.0 * KN, Mx=65 * KNm, My=45 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.CHARACTERISTIC
    )
    forces[3] = ForcesOnSection(
        id=3, Fz=50.0 * KN, Mx=90 * KNm, My=-90 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.CHARACTERISTIC
    )
    forces[4] = ForcesOnSection(
        id=4, Fz=80.0 * KN, Mx=80 * KNm, My=80 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.QUASI_PERMANENT
    )
    forces[5] = ForcesOnSection(
        id=5, Fz=30.0 * KN, Mx=80 * KNm, My=-80 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.QUASI_PERMANENT
    )
    forces[6] = ForcesOnSection(
        id=6, Fz=-30.0 * KN, Mx=-80 * KNm, My=80 * KNm,
        limitState=LimitState.SERVICEABILITY,
        frequency=Frequency.QUASI_PERMANENT
    )
    forces[7] = ForcesOnSection(
        id=7, Fz=80.0 * KN, Mx=80 * KNm, My=80 * KNm,
        limitState=LimitState.ULTIMATE
    )
    forces[8] = ForcesOnSection(
        id=8, Fz=30.0 * KN, Mx=80 * KNm, My=-80 * KNm,
        limitState=LimitState.ULTIMATE
    )
    forces[9] = ForcesOnSection(
        id=9, Fz=-30.0 * KN, Mx=-80 * KNm, My=80 * KNm,
        limitState=LimitState.ULTIMATE
    )
    forces[10] = ForcesOnSection(
        id=10, Fz=-30.0 * KN, Mx=0.0 * KNm, My=80 * KNm,
        limitState=LimitState.ULTIMATE
    )

    modeler.setCurrentModel("section_1")
    modeler.assignForces(forces)
    modeler.run(opt=Analysis.ELASTIC_SOLVER)
    modeler.run(opt=Analysis.CHECK_ELASTIC)
    modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)

    modeler.setCurrentModel("section_2")
    modeler.assignForces(forces)
    modeler.run(opt=Analysis.ELASTIC_SOLVER)
    modeler.run(opt=Analysis.CHECK_ELASTIC)
    modeler.run(opt=Analysis.CHECK_DOMAIN_SLU)

    input_model = modeler.exportModelInput()
    input_file = tmp_path.joinpath("test_007_input_model.json")
    input_file.write_text(input_model.model_dump_json(indent=4))

    output_model = modeler.getModelOutput()
    output_file = tmp_path.joinpath("test_007_output_model.json")
    output_file.write_text(output_model.model_dump_json(indent=4))

    modeler.plot(onlyWorst=False)


from pathlib import Path

import pytest
from pycivil.EXAStructural.lawcodes.codeNTC2018 import crackMeasure
from pycivil.EXAStructural.codes import Code
from pycivil.EXAStructural.materials import Concrete, ConcreteSteel
from pycivil.EXAStructural.templateRCRect import RCTemplRectEC2, SectionCrackedStates


# ----------------------
# BY-HAND TEST CASE #2.7
# ----------------------
def test_no_symmetric_M_unknown_001(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")

    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")

    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")

    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")

    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)

    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    rcSection.addSteelArea("LINE-MT", dist=76, d=24, nb=10, sd=49)
    rcSection.addSteelArea("LINE-MT", dist=1016, d=24, nb=10, sd=49)
    rcSection.addSteelArea("LINE-MT", dist=1070, d=24, nb=10, sd=49)
    rcSection.addSteelArea("LINE-MT", dist=1124, d=24, nb=10, sd=49)

    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = 0 * KN
    Med = 2400.0 / 3 * KNm
    assert rcSection.crackState() == SectionCrackedStates.UNKNOWN

    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() != SectionCrackedStates.UNKNOWN


# ----------------------
# BY-HAND TEST CASE #2.7
# ----------------------
def test_no_symmetric_M_decompressed_002(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")

    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")

    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")

    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")

    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)

    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    lineRebars01 = rcSection.addSteelArea("LINE-MT", dist=76, d=24, nb=10, sd=49)
    lineRebars02 = rcSection.addSteelArea("LINE-MT", dist=1016, d=24, nb=10, sd=49)
    lineRebars03 = rcSection.addSteelArea("LINE-MT", dist=1070, d=24, nb=10, sd=49)
    lineRebars04 = rcSection.addSteelArea("LINE-MT", dist=1124, d=24, nb=10, sd=49)

    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = 0 * KN
    Med = 2400.0 / 3 * KNm
    sigmac, sigmas, xi = rcSection.solverSLS_NM(Ned, Med, uncracked=True)
    assert pytest.approx([sigmac, sigmas, xi], rel=1e-2) == [
        -8.84 / 3,
        +96.97 / 3,
        649.39,
    ]

    for idx, i in enumerate(rcSection.getConcrStress()):
        if idx < 2:
            # BL and BR
            assert pytest.approx(i, rel=1e-3) == +7.4988 / 3
        else:
            # TL and TR
            assert pytest.approx(i, rel=1e-3) == -8.840 / 3

    for idx, i in enumerate(rcSection.getSteelStress()):
        stressExpected = 0
        if idx in lineRebars01:
            stressExpected = -117.11 / 3
        if idx in lineRebars02:
            stressExpected = +74.91 / 3
        if idx in lineRebars03:
            stressExpected = +85.94 / 3
        if idx in lineRebars04:
            stressExpected = +96.97 / 3
        assert pytest.approx(i, rel=1e-2) == stressExpected

    assert pytest.approx(rcSection.getConcreteMaterial().get_fctm(), rel=1e-2) == 3.02
    assert (
        pytest.approx(rcSection.getConcreteMaterial().get_fct_crack(), rel=1e-2) == 2.51
    )

    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() == SectionCrackedStates.DECOMPRESSED


# ----------------------
# BY-HAND TEST CASE #2.3
# ----------------------
def test_no_symmetric_M_cracked_bot_003(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")

    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")

    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")

    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")

    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)

    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    lineRebars01 = rcSection.addSteelArea("LINE-MT", dist=76, d=24, nb=10, sd=49)
    lineRebars02 = rcSection.addSteelArea("LINE-MT", dist=1016, d=24, nb=10, sd=49)
    lineRebars03 = rcSection.addSteelArea("LINE-MT", dist=1070, d=24, nb=10, sd=49)
    lineRebars04 = rcSection.addSteelArea("LINE-MT", dist=1124, d=24, nb=10, sd=49)

    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = 0 * KN
    Med = 2400.0 * KNm

    sigmac, sigmas, xi = rcSection.solverSLS_NM(Ned, Med, uncracked=False)

    assert pytest.approx([sigmac, sigmas, xi], rel=1e-3) == [-10.31, +207.10, 480.51]

    for idx, i in enumerate(rcSection.getConcrStress()):
        if idx < 2:
            # BL and BR
            assert pytest.approx(i, rel=1e-3) == 0.0
        else:
            # TL and TR
            assert pytest.approx(i, rel=1e-2) == -10.31

    for idx, i in enumerate(rcSection.getSteelStress()):
        stressExpected = 0
        if idx in lineRebars01:
            stressExpected = -130.19
        if idx in lineRebars02:
            stressExpected = +172.34
        if idx in lineRebars03:
            stressExpected = +189.72
        if idx in lineRebars04:
            stressExpected = +207.10
        assert pytest.approx(i, rel=1e-2) == stressExpected

    assert pytest.approx(rcSection.getConcreteMaterial().get_fctm(), rel=1e-2) == 3.02
    assert (
        pytest.approx(rcSection.getConcreteMaterial().get_fct_crack(), rel=1e-2) == 2.51
    )
    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() == SectionCrackedStates.CRACKED_BOT
    print("Crack parameters for BOT")
    print(rcSection.crackParam()[0])
    print("Crack parameters for TOP")
    print(rcSection.crackParam()[1])
    assert pytest.approx(rcSection.crackParam()[0].epsi, rel=1e-2) == +1.10869e-3
    assert pytest.approx(rcSection.crackParam()[0].coverInMaxSteel, rel=1e-2) == +64
    assert pytest.approx(rcSection.crackParam()[0].rebarDistance, rel=1e-2) == +25
    assert pytest.approx(rcSection.crackParam()[0].deq, rel=1e-2) == +24

    assert pytest.approx(rcSection.crackParam()[0].dgs, rel=1e-2) == +1070
    assert pytest.approx(rcSection.crackParam()[0].hcEff, rel=1e-2) == +239.73
    assert pytest.approx(rcSection.crackParam()[0].steelArea, rel=1e-2) == +13.572e3
    assert pytest.approx(rcSection.crackParam()[0].sigmasMax, rel=1e-2) == +207.2

    assert pytest.approx(rcSection.crackParam()[1].epsi, rel=1e-2) == -736.429e-6
    assert rcSection.crackParam()[1].coverInMaxSteel is None
    assert rcSection.crackParam()[1].rebarDistance is None
    assert rcSection.crackParam()[1].deq is None

    assert rcSection.crackParam()[1].dgs is None
    assert rcSection.crackParam()[1].hcEff is None
    assert rcSection.crackParam()[1].steelArea is None
    assert rcSection.crackParam()[1].sigmasMax is None

    crackOut = crackMeasure(
        epsiBot=rcSection.crackParam()[0].epsi,
        epsiTop=rcSection.crackParam()[1].epsi,
        deq=rcSection.crackParam()[0].deq,
        As=rcSection.crackParam()[0].steelArea,
        rebarsCover=rcSection.crackParam()[0].coverInMaxSteel,
        rebarsDistance=rcSection.crackParam()[0].rebarDistance,
        hcEff=rcSection.crackParam()[0].hcEff,
        beff=rcSection.getDimW(),
        hsec=rcSection.getDimH(),
        xi=rcSection.xi(),
        fck=rcSection.getConcreteMaterial().get_fck(),
        sigmas=rcSection.crackParam()[0].sigmasMax,
        Es=rcSection.getSteelMaterial().get_Es(),
        load="long",
    )

    print(crackOut.toDict())
    assert pytest.approx(crackOut.roeff, rel=1e-4) == +70.767e-3
    assert pytest.approx(crackOut.epsism, rel=1e-3) == +869.162e-6
    assert pytest.approx(crackOut.sigmas_stiffning, rel=1e-2) == +24.678
    sigmasMax = rcSection.crackParam()[0].sigmasMax
    assert crackOut.sigmas_stiffning <= 0.4 * sigmasMax
    assert pytest.approx(crackOut.deltasm, rel=1e-3) == +161.911
    assert pytest.approx(crackOut.deltasm1, rel=1e-4) == +161.911
    assert pytest.approx(crackOut.deltasm2, rel=1e-3) == +539.62
    assert crackOut.deltasm == crackOut.deltasm1
    assert pytest.approx(crackOut.zoneC, rel=1e-4) == +380.00
    assert pytest.approx(crackOut.wk, rel=1e-2) == +0.23923
    assert pytest.approx(crackOut.k2, rel=1e-10) == +0.5
    assert pytest.approx(crackOut.alpham, rel=1e-4) == +6.298


# ----------------------
# BY-HAND TEST CASE #2.3
# ----------------------
# As for test_no_symmetric_M_cracked_bot_003 mirroring section and forces
# produces cracked for top.
def test_no_symmetric_M_cracked_top_004(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")
    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")
    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")
    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")
    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)
    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    lineRebars01 = rcSection.addSteelArea("LINE-MT", dist=1200 - 76, d=24, nb=10, sd=49)
    lineRebars02 = rcSection.addSteelArea(
        "LINE-MT", dist=1200 - 1016, d=24, nb=10, sd=49
    )
    lineRebars03 = rcSection.addSteelArea(
        "LINE-MT", dist=1200 - 1070, d=24, nb=10, sd=49
    )
    lineRebars04 = rcSection.addSteelArea(
        "LINE-MT", dist=1200 - 1124, d=24, nb=10, sd=49
    )
    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = 0.0 * KN
    Med = -2400.0 * KNm

    sigmac, sigmas, xi = rcSection.solverSLS_NM(Ned, Med, uncracked=False)

    assert pytest.approx([sigmac, sigmas, xi], rel=1e-2) == [
        -10.31,
        +207.10,
        1200 - 480.51,
    ]
    for idx, i in enumerate(rcSection.getConcrStress()):
        if idx < 2:
            # BL and BR
            assert pytest.approx(i, rel=1e-3) == -10.31
        else:
            # TL and TR
            assert pytest.approx(i, rel=1e-3) == 0.0

    for idx, i in enumerate(rcSection.getSteelStress()):
        stressExpected = 0
        if idx in lineRebars01:
            stressExpected = -130.19
        if idx in lineRebars02:
            stressExpected = +172.34
        if idx in lineRebars03:
            stressExpected = +189.72
        if idx in lineRebars04:
            stressExpected = +207.10
        assert pytest.approx(i, rel=1e-2) == stressExpected

    assert pytest.approx(rcSection.getConcreteMaterial().get_fctm(), rel=1e-2) == 3.02
    assert (
        pytest.approx(rcSection.getConcreteMaterial().get_fct_crack(), rel=1e-2) == 2.51
    )
    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() == SectionCrackedStates.CRACKED_TOP
    print("Crack parameters for BOT")
    print(rcSection.crackParam()[0])
    print("Crack parameters for TOP")
    print(rcSection.crackParam()[1])

    assert pytest.approx(rcSection.crackParam()[1].epsi, rel=1e-2) == +1.10869e-3
    assert pytest.approx(rcSection.crackParam()[1].coverInMaxSteel, rel=1e-2) == +64
    assert pytest.approx(rcSection.crackParam()[1].rebarDistance, rel=1e-2) == +25
    assert pytest.approx(rcSection.crackParam()[1].deq, rel=1e-2) == +24

    assert pytest.approx(rcSection.crackParam()[1].dgs, rel=1e-2) == +1070
    assert pytest.approx(rcSection.crackParam()[1].hcEff, rel=1e-2) == +239.73
    assert pytest.approx(rcSection.crackParam()[1].steelArea, rel=1e-2) == +13.572e3
    assert pytest.approx(rcSection.crackParam()[1].sigmasMax, rel=1e-2) == +207.2

    assert pytest.approx(rcSection.crackParam()[0].epsi, rel=1e-2) == -736.429e-6
    assert rcSection.crackParam()[0].coverInMaxSteel is None
    assert rcSection.crackParam()[0].rebarDistance is None
    assert rcSection.crackParam()[0].deq is None

    assert rcSection.crackParam()[0].dgs is None
    assert rcSection.crackParam()[0].hcEff is None
    assert rcSection.crackParam()[0].steelArea is None
    assert rcSection.crackParam()[0].sigmasMax is None

    crackOut = crackMeasure(
        epsiBot=rcSection.crackParam()[0].epsi,
        epsiTop=rcSection.crackParam()[1].epsi,
        deq=rcSection.crackParam()[1].deq,
        As=rcSection.crackParam()[1].steelArea,
        rebarsCover=rcSection.crackParam()[1].coverInMaxSteel,
        rebarsDistance=rcSection.crackParam()[1].rebarDistance,
        hcEff=rcSection.crackParam()[1].hcEff,
        beff=rcSection.getDimW(),
        hsec=rcSection.getDimH(),
        xi=rcSection.xi(),
        fck=rcSection.getConcreteMaterial().get_fck(),
        sigmas=rcSection.crackParam()[1].sigmasMax,
        Es=rcSection.getSteelMaterial().get_Es(),
        load="long",
    )

    print(crackOut.toDict())
    assert pytest.approx(crackOut.roeff, rel=1e-4) == +70.767e-3
    assert pytest.approx(crackOut.epsism, rel=1e-3) == +869.162e-6
    assert pytest.approx(crackOut.sigmas_stiffning, rel=1e-2) == +24.678
    sigmasMax = rcSection.crackParam()[1].sigmasMax
    assert crackOut.sigmas_stiffning <= 0.4 * sigmasMax
    assert pytest.approx(crackOut.deltasm, rel=1e-3) == +161.911
    assert pytest.approx(crackOut.deltasm1, rel=1e-4) == +161.911
    assert pytest.approx(crackOut.deltasm2, rel=1e-3) == +539.62
    assert crackOut.deltasm == crackOut.deltasm1
    assert pytest.approx(crackOut.zoneC, rel=1e-4) == +380.00
    assert pytest.approx(crackOut.wk, rel=1e-2) == +0.23923
    assert pytest.approx(crackOut.k2, rel=1e-10) == +0.5
    assert pytest.approx(crackOut.alpham, rel=1e-4) == +6.298


# ----------------------
# BY-HAND TEST CASE #2.9
# ----------------------
def test_no_symmetric_N_cracked_005(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")

    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")

    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")

    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")

    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)

    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    lineRebars01 = rcSection.addSteelArea("LINE-MT", dist=76, d=24, nb=10, sd=49)
    lineRebars02 = rcSection.addSteelArea("LINE-MT", dist=1016, d=24, nb=10, sd=49)
    lineRebars03 = rcSection.addSteelArea("LINE-MT", dist=1070, d=24, nb=10, sd=49)
    lineRebars04 = rcSection.addSteelArea("LINE-MT", dist=1124, d=24, nb=10, sd=49)

    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = +3619 * KN
    Med = 0.0 * KNm
    sigmac, sigmas, xi = rcSection.solverSLS_NM(Ned, Med, uncracked=False)
    assert pytest.approx([sigmac, sigmas, xi], rel=1e-2) == [0.0, +376.27, 1666.0]

    for idx, i in enumerate(rcSection.getConcrStress()):
        if idx < 2:
            # BL and BR
            assert pytest.approx(i, rel=1e-3) == 0.0
        else:
            # TL and TR
            assert pytest.approx(i, rel=1e-3) == 0.0

    for idx, i in enumerate(rcSection.getSteelStress()):
        stressExpected = 0
        if idx in lineRebars01:
            stressExpected = 376.27
        if idx in lineRebars02:
            stressExpected = 153.86
        if idx in lineRebars03:
            stressExpected = 141.05
        if idx in lineRebars04:
            stressExpected = 128.27
        assert pytest.approx(i, rel=1e-2) == stressExpected

    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() == SectionCrackedStates.CRACKED

    print("Crack parameters for TOP AND BOTTOM")
    print("-------------- BOT ----------------")
    print(rcSection.crackParam()[0])
    print("-------------- TOP ----------------")
    print(rcSection.crackParam()[1])

    assert pytest.approx(rcSection.crackParam()[0].epsi, rel=1e-2) == +525.118e-6
    assert pytest.approx(rcSection.crackParam()[0].coverInMaxSteel, rel=1e-2) == +64
    assert pytest.approx(rcSection.crackParam()[0].rebarDistance, rel=1e-2) == +25
    assert pytest.approx(rcSection.crackParam()[0].deq, rel=1e-2) == +24

    assert pytest.approx(rcSection.crackParam()[0].dgs, rel=1e-2) == +1070
    assert pytest.approx(rcSection.crackParam()[0].hcEff, rel=1e-2) == +325.0
    assert pytest.approx(rcSection.crackParam()[0].steelArea, rel=1e-2) == +13572
    assert pytest.approx(rcSection.crackParam()[0].sigmasMax, rel=1e-2) == +153.86

    assert pytest.approx(rcSection.crackParam()[1].epsi, rel=1e-2) == +1.877e-3
    assert pytest.approx(rcSection.crackParam()[1].coverInMaxSteel, rel=1e-2) == +64
    assert pytest.approx(rcSection.crackParam()[1].rebarDistance, rel=1e-2) == +25
    assert pytest.approx(rcSection.crackParam()[1].deq, rel=1e-2) == +24

    assert pytest.approx(rcSection.crackParam()[1].dgs, rel=1e-2) == +1124
    assert pytest.approx(rcSection.crackParam()[1].hcEff, rel=1e-2) == +190.0
    assert pytest.approx(rcSection.crackParam()[1].steelArea, rel=1e-2) == +4523.893
    assert pytest.approx(rcSection.crackParam()[1].sigmasMax, rel=1e-2) == +376.27

    crackOutTop = crackMeasure(
        epsiBot=rcSection.crackParam()[0].epsi,
        epsiTop=rcSection.crackParam()[1].epsi,
        deq=rcSection.crackParam()[1].deq,
        As=rcSection.crackParam()[1].steelArea,
        rebarsCover=rcSection.crackParam()[1].coverInMaxSteel,
        rebarsDistance=rcSection.crackParam()[1].rebarDistance,
        hcEff=rcSection.crackParam()[1].hcEff,
        beff=rcSection.getDimW(),
        hsec=rcSection.getDimH(),
        xi=rcSection.xi(),
        fck=rcSection.getConcreteMaterial().get_fck(),
        sigmas=rcSection.crackParam()[1].sigmasMax,
        Es=rcSection.getSteelMaterial().get_Es(),
        load="long",
    )

    print(crackOutTop.toDict())
    assert pytest.approx(crackOutTop.roeff, rel=1e-3) == +29.757e-3
    assert pytest.approx(crackOutTop.epsism, rel=1e-2) == +1.562e-3
    assert pytest.approx(crackOutTop.sigmas_stiffning, rel=1e-2) == +48.203
    assert pytest.approx(crackOutTop.deltasm, rel=1e-2) == +231.22
    assert pytest.approx(crackOutTop.deltasm1, rel=1e-2) == +231.22
    assert pytest.approx(crackOutTop.deltasm2, rel=1e-3) == +1249.5
    assert crackOutTop.deltasm == crackOutTop.deltasm1
    assert pytest.approx(crackOutTop.zoneC, rel=1e-4) == +380.00
    assert pytest.approx(crackOutTop.wk, rel=1e-2) == +0.61398
    assert pytest.approx(crackOutTop.k2, rel=1e-3) == +0.6399
    assert pytest.approx(crackOutTop.alpham, rel=1e-4) == +6.298

    crackOutBot = crackMeasure(
        epsiBot=rcSection.crackParam()[0].epsi,
        epsiTop=rcSection.crackParam()[1].epsi,
        deq=rcSection.crackParam()[0].deq,
        As=rcSection.crackParam()[0].steelArea,
        rebarsCover=rcSection.crackParam()[0].coverInMaxSteel,
        rebarsDistance=rcSection.crackParam()[0].rebarDistance,
        hcEff=rcSection.crackParam()[0].hcEff,
        beff=rcSection.getDimW(),
        hsec=rcSection.getDimH(),
        xi=rcSection.xi(),
        fck=rcSection.getConcreteMaterial().get_fck(),
        sigmas=rcSection.crackParam()[0].sigmasMax,
        Es=rcSection.getSteelMaterial().get_Es(),
        load="long",
    )

    print(crackOutBot.toDict())
    assert pytest.approx(crackOutBot.roeff, rel=1e-4) == +52.2e-3
    assert pytest.approx(crackOutBot.sigmas_stiffning, rel=1e-2) == +30.75
    assert pytest.approx(crackOutBot.epsism, rel=1e-3) == +586.238e-6
    assert pytest.approx(crackOutBot.deltasm, rel=1e-3) == +186.84
    assert pytest.approx(crackOutBot.deltasm1, rel=1e-3) == +186.84
    assert pytest.approx(crackOutBot.deltasm2, rel=1e-3) == +1249.5
    assert crackOutTop.deltasm == crackOutTop.deltasm1
    assert pytest.approx(crackOutTop.zoneC, rel=1e-4) == +380.00
    assert pytest.approx(crackOutBot.wk, rel=1e-2) == +0.186205
    assert pytest.approx(crackOutBot.k2, rel=1e-3) == +0.6399
    assert pytest.approx(crackOutBot.alpham, rel=1e-4) == +6.298


# ----------------------
# BY-HAND TEST CASE #2.6
# ----------------------
def test_cracked_compressed_NM_006(tmp_path: Path):
    # Setting code for check
    code = Code("NTC2008")

    concrete = Concrete(descr="My concrete")
    concrete.setByCode(code, "C32/40")

    # Setting code for check
    steel = ConcreteSteel(descr="My steel")
    steel.setByCode(code, "B450C")

    # Build checkable structural system
    rcSection = RCTemplRectEC2(1, "Template RC Section")

    # Concrete dimension
    rcSection.setDimH(1200)
    rcSection.setDimW(800)

    # Longitudinal rebars
    # Aggregates diameter maximum 25mm
    lineRebars01 = rcSection.addSteelArea("LINE-MT", dist=76, d=24, nb=10, sd=49)
    lineRebars02 = rcSection.addSteelArea("LINE-MT", dist=1016, d=24, nb=10, sd=49)
    lineRebars03 = rcSection.addSteelArea("LINE-MT", dist=1070, d=24, nb=10, sd=49)
    lineRebars04 = rcSection.addSteelArea("LINE-MT", dist=1124, d=24, nb=10, sd=49)

    # Setting units
    KN = 1000
    KNm = 1000 * 1000
    Ned = -3619.0 * KN
    Med = 0.0 * KNm
    sigmac, sigmas, xi = rcSection.solverSLS_NM(Ned, Med, uncracked=False)
    assert pytest.approx([sigmac, sigmas, xi], rel=1e-2) == [-3.599, -36.92, 3557.496]

    for idx, i in enumerate(rcSection.getConcrStress()):
        if idx < 2:
            # BL and BR
            assert pytest.approx(i, rel=1e-3) == -2.385
        else:
            # TL and TR
            assert pytest.approx(i, rel=1e-2) == -3.599

    for idx, i in enumerate(rcSection.getSteelStress()):
        stressExpected = 0
        if idx in lineRebars01:
            stressExpected = -52.84
        if idx in lineRebars02:
            stressExpected = -38.56
        if idx in lineRebars03:
            stressExpected = -37.74
        if idx in lineRebars04:
            stressExpected = -36.92
        assert pytest.approx(i, rel=1e-2) == stressExpected

    rcSection.solverCrack(Ned, Med)
    assert rcSection.crackState() == SectionCrackedStates.COMPRESSED



"""A package with tools built around pycivil and other python libraries

The package contains tools like post-processors (for Strand7, Midas etc.),
checkers (for standard codes like Eurocodes, Italian Codes) and wrappers
around important python packages like vtk or gmsh.
"""


import math
import os
from dataclasses import field
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

from pydantic import BaseModel, ConfigDict

from pycivil.EXAUtils.EXAExceptions import EXAExceptions as Ex
from pycivil.EXAGeometry.clouds import PointCloud2d, PointCloud2dMdl
from pycivil.EXAGeometry.geometry import Node2d, Point2d
from pycivil.EXAGeometry.shapes import Shape, ShapeCircle, ShapeRect, ShapesEnum
from pycivil.EXAStructural.loads import ForcesOnSection as Forces
from pycivil.EXAStructural.loads import Frequency_Enum as Frequency
from pycivil.EXAStructural.loads import LimiteState_Enum as LimitState
from pycivil.EXAStructural.materials import (
    Concrete,
    ConcreteModel,
    ConcreteSteel,
    SteelModel,
)
from pycivil.EXAStructural.modeler import SectionModeler
from pycivil.EXAStructural.plot import Geometry2DPlot, SectionPlot, SectionPlotView_Enum
from pycivil.EXAStructural.sections import RectangularShape, SectionStates, ShapeEnum as SectionShapeTp, TShape
from pycivil.EXAUtils.logging import log
from pycivil.EXAUtils.solver import Solver

from pycivil.EXAUtils.report import (
    FragmentsBuilder,
    Fragment,
    getTemplatesPath,
    EnumFBSection
)

from pycivil.EXAUtils.latexReportMakers import (
    CodesFB,
    ConcreteFB,
    SteelConcreteFB,
    ForcesOnSectionListFB
)


class Analysis(str, Enum):
    model_config = ConfigDict(use_enum_values=True)
    CHECK_DOMAIN_SLU = "CHECK_DOMAIN_SLU"
    BOUNDING_SLU = "BOUNDING_SLU"
    ELASTIC_SOLVER = "ELASTIC_SOLVER"
    CHECK_ELASTIC = "CHECK_ELASTIC"


class ArtifactType(str, Enum):
    model_config = ConfigDict(use_enum_values=True)
    SECTION_GEOMETRY = "SECTION_GEOMETRY"
    DOMAIN_SLU = "DOMAIN_SLU"
    ELASTIC_STRESS = "ELASTIC_STRESS"
    REPORT_NOT_EDITABLE = "REPORT_NOT_EDITABLE"
    REPORT_EDITABLE = "REPORT_EDITABLE"
    UNKNOWN = "UNKNOWN"


class Artifact(BaseModel):
    tp: ArtifactType = ArtifactType.UNKNOWN
    path: Path | str = Path()
    name: str = "unknown_artifact"
    ext: str = ""
    def file_name(self) -> Path:
        return self.path / Path(self.name).with_suffix("."+self.ext)

class DomainSLUOptions(BaseModel):
    degreeDivision: int = 15
    rationDivision: int = 30
    rebuild: bool = False


class BoundingSLUOptions(BaseModel):
    degreeDivision: int = 15
    rationDivision: int = 30
    rebuild: bool = False


class CheckResults(BaseModel):
    check: bool = False
    safetyFactor: Union[float, Literal["inf"]] = 0.0


class DomainSLUCheck(BaseModel):
    interactionDomain: CheckResults = CheckResults()
    loadId: int


class InteractionPoints2d(BaseModel):
    component: Literal["Fz", "Mx", "My"] = "Fz"
    pointsCloud2d: PointCloud2dMdl = PointCloud2dMdl()


class DomainSLUCheckLog(BaseModel):
    logs: List[str] = field(default_factory=list)
    Nmin: float
    Nmax: float
    Nrd: Union[float, None]
    Ned: float
    Mxrd: Union[float, None]
    Myrd: Union[float, None]
    Mxed: float
    Myed: float
    loadId: int
    check: bool
    domain: InteractionPoints2d = InteractionPoints2d()
    artifacts: List[Artifact] = []

class CheckDomainSLUResults(BaseModel, FragmentsBuilder):
    logs: List[str] = field(default_factory=list[str])
    globalCheck: Optional[CheckResults] = None
    loadIdCheck: Optional[int] = None
    loadsCheck: Dict[int, DomainSLUCheck] = field(
        default_factory=dict[int, DomainSLUCheck]
    )
    loadsCheckLogs: Dict[int, DomainSLUCheckLog] = field(
        default_factory=dict[int, DomainSLUCheckLog]
    )
    artifacts: List[Artifact] = []
    def model_post_init(self, context: Any, /) -> None:
        self._setFragmentOptions({"job_path": Path()})

    def buildFragment(self) -> Fragment:
        f = Fragment(getTemplatesPath())
        place_holders = dict()
        place_holders["check_SLU_MN"] = []
        place_holders["domainUrls"] = []
        artifacts = []
        for key_load, check_loads in self.loadsCheckLogs.items():
            place_holders_check = dict()
            place_holders_check["id"] = key_load
            place_holders_check["Ned"] = f"{check_loads.Ned / 1e3:.1f}"
            place_holders_check["Mxed"] = f"{check_loads.Mxed / 1e6:.1f}"
            place_holders_check["Myed"] = f"{check_loads.Myed / 1e6:.1f}"

            if check_loads.Nrd is None:
                place_holders_check["Nrd"] = f"n.d."
            else:
                place_holders_check["Nrd"] = f"{check_loads.Nrd / 1e3:.1f}"

            if check_loads.Mxrd is None or check_loads.Mxed == 0.0:
                place_holders_check["Mxrd"] = f"n.d."
            else:
                place_holders_check["Mxrd"] = f"{check_loads.Mxrd / 1e6:.1f}"

            if check_loads.Myrd is None or check_loads.Myed == 0.0:
                place_holders_check["Myrd"] = f"n.d."
            else:
                place_holders_check["Myrd"] = f"{check_loads.Myrd / 1e6:.1f}"

            place_holders_check["FS"] = f"{self.loadsCheck[key_load].interactionDomain.safetyFactor :.2f}"
            if self.loadsCheck[key_load].interactionDomain.check:
                place_holders_check["check"] = f"OK"
            else:
                place_holders_check["check"] = f"NOOK"

            for artifact in check_loads.artifacts:
                if artifact.tp == ArtifactType.DOMAIN_SLU:
                    geometryFilePath_encoded_for_latex = str(str(self.getFragmentOptions()["job_path"] / artifact.file_name())).replace('\\', '/')
                    artifacts.append(
                        {
                            "domainUrl": geometryFilePath_encoded_for_latex,
                            "idLoad": key_load,
                            "newLine": False
                        }
                    )

            place_holders["check_SLU_MN"].append(place_holders_check)

        # TABLE for figures
        #
        maxCols = 2
        maxRows = 3
        max_elem_for_figure = maxCols * maxRows
        nb_elem = len(artifacts)
        nb_figs = math.ceil(nb_elem / max_elem_for_figure)

        figures_matrix = []
        elem = []
        fig_index = 0

        if nb_elem == 1:
            figure_width = 0.9
        else:
            figure_width = 0.41

        for a in artifacts:
            elem.append(a)
            if math.ceil(len(elem) / maxCols) == math.floor(len(elem) / maxCols):
                a["newLine"] = True
            if len(elem) == max_elem_for_figure:
                fig_index += 1
                figures_matrix.append({"domainUrls": elem, "figIndex": fig_index})
                elem = []
        if len(elem) < max_elem_for_figure:
            fig_index += 1
            figures_matrix.append({"domainUrls": elem, "figIndex": fig_index})

        place_holders["figures"] = figures_matrix
        place_holders["nbFigs"] = nb_figs
        place_holders["figuresWidth"] = figure_width
        f.add_template("template-ita-rc-gen-slu-nm.tex", place_holders)
        return f


class ElasticSLSCheck(BaseModel):
    elasticCheck: CheckResults = CheckResults()
    loadId: int = -1


class ElasticSolverResult(BaseModel):
    Ned: float = 0.0
    Mxed: float = 0.0
    Myed: float = 0.0
    sigmacMin: float = 0.0
    sigmacMax: float = 0.0
    sigmasMin: float = 0.0
    sigmasMax: float = 0.0
    tensionPlane: Tuple[float, float, float] = [0.0, 0.0, 0.0]
    sigmaci: dict[int, float] = field(default_factory=dict[int, float])
    sigmasi: dict[int, float] = field(default_factory=dict[int, float])
    state: SectionStates = SectionStates.UNKNOWN
    artifacts: List[Artifact] = []
    # maxIt: int = 0
    # tolIt: float = 0.0


class ElasticSolverResults(BaseModel):
    logs: List[str] = field(default_factory=list[str])
    elasticResults: Dict[int, ElasticSolverResult] = field(
        default_factory=dict[int, ElasticSolverResult]
    )


class ElasticSLSCheckLog(BaseModel):
    logs: List[str] = field(default_factory=list)
    sigmacMinQP: Union[float, None] = None
    sigmacQP: Union[float, None] = None
    sigmacMinCH: Union[float, None] = None
    sigmacCH: Union[float, None] = None
    sigmasMaxCH: Union[float, None] = None
    sigmasCH: Union[float, None] = None
    frequency: Frequency = Frequency.FREQUENCY_ND


class ElasticSLSCheckResults(BaseModel):
    logs: List[str] = field(default_factory=list[str])
    globalCheck: Optional[CheckResults] = None
    loadIdCheck: Optional[int] = None
    loadsCheck: dict[int, ElasticSLSCheck] = field(
        default_factory=dict[int, ElasticSLSCheck]
    )
    loadsCheckLogs: dict[int, ElasticSLSCheckLog] = field(
        default_factory=dict[int, ElasticSLSCheckLog]
    )
    artifacts: List[Artifact] = []


class BoundingSLUResults(BaseModel):
    logs: List[str] = field(default_factory=list)
    minForces: Optional[Forces] = None
    maxForces: Optional[Forces] = None


class Results(BaseModel):
    logs: List[str] = field(default_factory=list)
    sectionKey: str = ""
    domainSLU: Optional[CheckDomainSLUResults] = None
    boundingSLU: Optional[BoundingSLUResults] = None
    elasticSolver: Optional[ElasticSolverResults] = None
    elasticCheck: Optional[ElasticSLSCheckResults] = None
    artifacts: List[Artifact] = []


class RCGenSectionsOutput(BaseModel):
    sectionResults: Dict[int, Results] = field(default_factory=dict[int, Results])


class Options(BaseModel):
    domainSLU: Optional[DomainSLUOptions] = DomainSLUOptions()
    boundingSLU: Optional[BoundingSLUOptions] = BoundingSLUOptions()


class Node2dModel(BaseModel):
    id: int = 0
    coords: Tuple[float, float] = [0.0, 0.0]


class Triangle2dModel(BaseModel):
    id: Tuple[int, int, int] = [0, 0, 0]


class MeshedShape(BaseModel):
    vertices: Dict[int, Tuple[float, float]] = {}
    triangles: Dict[int, Tuple[int, int, int]] = {}
    namedVertices: Dict[str, int] = {}


class Rebar(BaseModel):
    idRebar: int = -1
    idVertex: int = -1
    radius: float = 0


class Geom(BaseModel):
    meshedConcrete: MeshedShape = MeshedShape()
    shapedConcrete: Optional[Union[RectangularShape, TShape]] = None
    rebars: Optional[Dict[int, Rebar]] = {}

class MatForSection(BaseModel):
    concrete: Optional[int] = None
    rebars: Optional[int] = None

class RCGenSection(BaseModel, FragmentsBuilder):
    key: str = ""
    geometry: Geom = Geom()
    materials: MatForSection = MatForSection()
    loads: Dict[int, Forces] = {}

    def model_post_init(self, context: Any, /) -> None:
        self._setFragmentOptions({"geometry_file_path": Path()})

    def buildFragment(self) -> Fragment:
        f = Fragment(getTemplatesPath())
        f.add_line(r"\subsection{Geometria}")

        rebars_place_holders = []
        vertices_place_holders = []
        named_vertices_place_holders = []
        triangles_place_holders = []
        place_holders = {
            "vertices": vertices_place_holders,
            "triangles": triangles_place_holders,
            "rebars": rebars_place_holders,
            "geometryFigure": False,
            "geometryUrl": ""
        }

        # Build figure with few tricks for latex
        #
        geometryFilePath = self.getFragmentOptions()["geometry_file_path"]
        if not isinstance(geometryFilePath, Path):
            log("WRN", f"Fragment option *geometry_file_path* isn't Path class", 2)
        if not geometryFilePath.exists():
            log("WRN", f"FIle path {geometryFilePath} doesn't exists !!!", 2)
        else:
            place_holders["geometryFigure"] = True
        # Only for latex also in windows path repr we need to
        # use /
        geometryFilePath_encoded_for_latex = str(geometryFilePath).replace('\\', '/')
        place_holders["geometryUrl"] = geometryFilePath_encoded_for_latex

        # TABLE for rebars
        #
        rebar_vertex_id = []
        for key_r, value_r in self.geometry.rebars.items():
            rebar_vertex_id.append(value_r.idVertex)
            rebars_place_holders.append(
                {
                    "id": key_r,
                    "xPos": f"{self.geometry.meshedConcrete.vertices[value_r.idVertex][0]:.1f}",
                    "yPos": f"{self.geometry.meshedConcrete.vertices[value_r.idVertex][1]:.1f}",
                    "diam": f"{value_r.radius*2:.0f}",
                    "id_v": f"{rebar_vertex_id[-1]:.0f}",
                }
            )

        # TABLE for solid vertices
        #
        for key_v, value_v in self.geometry.meshedConcrete.vertices.items():
            if key_v not in rebar_vertex_id:
                vertices_place_holders.append(
                    {
                        "id": key_v,
                        "xPos": f"{value_v[0]:.1f}",
                        "yPos": f"{value_v[1]:.1f}"
                    }
                )

        # TABLE for named vertices
        #
        if len(self.geometry.meshedConcrete.namedVertices) > 0:
            for key_n, value_n in self.geometry.meshedConcrete.namedVertices.items():
                named_vertices_place_holders.append(
                    {
                        "name": key_n,
                        "xPos": f"{self.geometry.meshedConcrete.vertices[value_n][0]:.1f}",
                        "yPos": f"{self.geometry.meshedConcrete.vertices[value_n][1]:.1f}"
                    }
                )
            place_holders["namedVertices"] = named_vertices_place_holders
        if self.geometry.shapedConcrete is not None:
            if isinstance(self.geometry.shapedConcrete, RectangularShape):
                place_holders["sectionShape"] = "rettangolare"
                place_holders["rectangularShape"] = True
                place_holders["shape_width"] = f"{self.geometry.shapedConcrete.width:.1f}"
                place_holders["shape_height"] = f"{self.geometry.shapedConcrete.height:.1f}"
            elif isinstance(self.geometry.shapedConcrete, TShape):
                log("WRN", f"Shape TShape not implemented !!!", 2)
            else:
                log("WRN", f"Shape unknown !!!", 2)

        # TABLE for triangles
        #
        for key_t, value_t in self.geometry.meshedConcrete.triangles.items():
            triangles_place_holders.append(
                {
                    "id": key_t,
                    "id1": f"{value_t[0]:.0f}",
                    "id2": f"{value_t[1]:.0f}",
                    "id3": f"{value_t[2]:.0f}",
                }
            )
        f.add_template("template-ita-rc-section-gen.tex", place_holders)
        return f


class MatCatalogue(BaseModel):
    concrete: Dict[int, ConcreteModel] = {}
    steel: Dict[int, SteelModel] = {}


class IdsManagement(BaseModel):
    lastSectionId: int = 0
    lastTriangleId: int = 0
    lastNodeId: int = 0
    lastMatId: int = 0
    lastRebarId: int = 0


class CommandType(str, Enum):
    RUN = "RUN"
    PLOT = "PLOT"
    NONE = "NONE"


class CommandArg(BaseModel):
    name: str = ""
    value: Optional[Any] = None


class Command(BaseModel):
    model_config = ConfigDict(use_enum_values=True)
    type: CommandType = CommandType.NONE
    args: List[CommandArg] = []


class RCGenSectionsInput(BaseModel):
    matCatalogue: MatCatalogue = MatCatalogue()
    sections: Dict[int, RCGenSection] = {}
    idsMng: IdsManagement = IdsManagement()
    commands: Optional[List[Command]] = None


class RCGenSectionsModeler(Solver):
    def __init__(self):
        super().__init__()
        self._setSolverName("RCGen")

        # Logger level
        self.__ll: Literal[0, 1, 2, 3] = 1

        # Modeler object
        self.__md = SectionModeler()
        self.__md.setLogLevel(self.__ll)

        # Id management
        self.__last_section_id = 0
        self.__last_triangle_id = 0
        self.__last_node_id = 0
        self.__last_mat_id = 0
        self.__last_rebar_id = 0

        # Section model keys to id for modeler
        self.__sectionKeys: Dict[str, int] = {}

        # Options for analysis
        self.__options: Options = Options()

        self.__journalOfCommands: List[Command] = []

        # Results for analisys
        self.__results: RCGenSectionsOutput = RCGenSectionsOutput()

        # Material used overall with id
        self.__concreteMaterial: Dict[int, Concrete] = {}
        self.__rebarsMaterial: Dict[int, ConcreteSteel] = {}

        # Material on sections with id for model and id for material assigned
        self.__matForConcrete: Dict[int, int] = {}
        self.__matForRebars: Dict[int, int] = {}

        # Loads on sections
        self.__forces: Dict[int, Dict[int, Forces]] = {}

        # Elements for shaped section
        self.__shapeDicts: Dict[int, Shape] = {}
        self.__namedVerticesDict: Dict[int, Dict[str, int]] = {}

        self.__trianglesIds: Union[Dict[int, List[int]], None] = None

    @property
    def logLevel(self):
        return self.__ll

    @logLevel.setter
    def logLevel(self, value):
        self.__ll = value
        self.__md.setLogLevel(value)

    def setJobPath(self, path: str | Path) -> bool:
        if isinstance(path, Path):
            path = str(path)
        log("INF", f"Setting job path {path}...", self.__ll)
        if Path(path).is_dir() and Path(path).exists():
            log("INF", f"... path exist, is a dir and will be set.", self.__ll)
            return super().setJobPath(path)
        log("ERR", f"... path doesn't exist or isn't a dir !!!", self.__ll)
        return False

    def _buildSolverFromModelInput(self, model: BaseModel) -> bool:
        log("INF", f"Model validation ...", self.__ll)
        if not isinstance(model,RCGenSectionsInput):
            log("ERR",f"Model must be RCGenSectionsInput class !!!", self.__ll)
            return False
        log("INF", f"... model is RCGenSectionsInput class", self.__ll)

        log("INF", f"Build catalogue for concrete ...", self.__ll)
        for k in model.matCatalogue.concrete.keys():
            self.addConcreteLaw(mat=model.matCatalogue.concrete[k].toMaterial(), idm=k)

        log("INF", f"Build catalogue for rebars ...", self.__ll)
        for k in model.matCatalogue.steel.keys():
            self.addRebarLaw(mat=model.matCatalogue.steel[k].toMaterial(), idm=k)

        # Sections have key and ids unique
        #
        log("INF", f"Build sections ...", self.__ll)
        for idx, key in enumerate(model.sections.keys()):
            log("INF", f"... adding section id={key} key={model.sections[key].key}", self.__ll)
            self.addSectionModel(key=model.sections[key].key, ids=key)
            if model.sections[key].geometry.shapedConcrete is not None:
                log("INF", f"... forming geometry shaped ...", self.__ll)
                if model.sections[key].geometry.shapedConcrete.tp == SectionShapeTp.RECT:
                    log("INF", f"...... rectangular shape ...", self.__ll)
                    w = model.sections[key].geometry.shapedConcrete.width
                    h = model.sections[key].geometry.shapedConcrete.height
                    if self.__shapeDicts is None:
                        self.__shapeDicts = {self.__md.getCurrent(): ShapeRect(w, h)}
                    else:
                        self.__shapeDicts[self.__md.getCurrent()] = ShapeRect(w, h)
                    if self.__namedVerticesDict is None:
                        self.__namedVerticesDict = {
                            self.__md.getCurrent():
                            model.sections[key].geometry.meshedConcrete.namedVertices
                        }
                    else:
                        self.__namedVerticesDict[self.__md.getCurrent()] = \
                            model.sections[key].geometry.meshedConcrete.namedVertices
                else:
                    log("ERR", f"...... shaped unknown. Exit with False !!!", self.__ll)
                    return False
            else:
                log("INF", f"... adding section id={key} key={model.sections[key].key}", self.__ll)
                log("INF", f"... forming geometry with vertex ...", self.__ll)

            log("INF", f"... adding vertices ...", self.__ll)
            vertices = model.sections[key].geometry.meshedConcrete.vertices
            for key_v in vertices.keys():
                self.addNode(Node2d(vertices[key_v][0], vertices[key_v][1], key_v))

            log("INF", f"... adding triangles ...", self.__ll)
            triangles = model.sections[key].geometry.meshedConcrete.triangles
            for key_t in triangles.keys():
                self.addTriangle(triangles[key_t][0], triangles[key_t][1], triangles[key_t][2], key_t)

            log("INF", f"... adding rebars ...", self.__ll)
            for key_r in model.sections[key].geometry.rebars.keys():
                rebar = model.sections[key].geometry.rebars[key_r]
                self.addRebar(idr=rebar.idRebar, idn=rebar.idVertex, diameter=rebar.radius * 2, use_vertex=True)

            log("INF", f"... assign material laws ...", self.__ll)
            self.assignConcreteLawToCurrentModel(model.sections[key].materials.concrete)
            self.assignRebarLawToCurrentModel(model.sections[key].materials.rebars)

            log("INF", f"... adding loads ...", self.__ll)
            self.assignForces(model.sections[key].loads)

        log("INF", f"... assign last ids ...", self.__ll)
        self.__last_section_id = model.idsMng.lastSectionId
        self.__last_triangle_id = model.idsMng.lastTriangleId
        self.__last_node_id = model.idsMng.lastNodeId
        self.__last_mat_id = model.idsMng.lastMatId
        self.__last_rebar_id = model.idsMng.lastRebarId

        log("INF", f"... journal of commands ...", self.__ll)
        self.__journalOfCommands = model.commands
        return True

    def _buildModelInputFromSolver(self) -> BaseModel:
        modelInput = RCGenSectionsInput()
        # ----------------------------------------------------------------------
        #                      I D S   M A N A G E M E N T
        # ----------------------------------------------------------------------
        modelInput.idsMng.lastSectionId = self.__last_section_id
        modelInput.idsMng.lastTriangleId = self.__last_triangle_id
        modelInput.idsMng.lastNodeId = self.__last_node_id
        modelInput.idsMng.lastMatId = self.__last_mat_id
        modelInput.idsMng.lastRebarId = self.__last_rebar_id
        # ----------------------------------------------------------------------
        #               J O U R N A L   O F   C O M M A N D S
        # ----------------------------------------------------------------------
        modelInput.commands = self.__journalOfCommands
        # ----------------------------------------------------------------------
        #              M A T E R I A L S   C A T A L O G U E S
        # ----------------------------------------------------------------------
        # Concrete
        #
        for k in self.__concreteMaterial.keys():
            modelInput.matCatalogue.concrete[k] = ConcreteModel()
            modelInput.matCatalogue.concrete[k].fromMaterial(self.__concreteMaterial[k])

        # Steel
        #
        for k in self.__rebarsMaterial.keys():
            modelInput.matCatalogue.steel[k] = SteelModel()
            modelInput.matCatalogue.steel[k].fromMaterial(self.__rebarsMaterial[k])

        for key_section in self.__sectionKeys.keys():
            id_section = self.__sectionKeys[key_section]
            sectionModel = RCGenSection(key=key_section)
            # ------------------------------------------------------------------
            #                        G E O M E T R Y
            # ------------------------------------------------------------------
            # Shaped section datas
            #
            if self.__shapeDicts is not None:
                if id_section in self.__shapeDicts.keys():
                    shape = self.__shapeDicts[id_section]
                    if isinstance(shape, ShapeRect):
                        sectionModel.geometry.shapedConcrete = RectangularShape(
                            id=shape.getIds(),
                            descr=shape.getDesc(),
                            width=shape.w(),
                            height=shape.h(),
                        )
                        sectionModel.geometry.meshedConcrete.namedVertices = (
                            self.__namedVerticesDict[id_section]
                        )

            # Meshed section datas
            #
            self.__md.setCurrent(id_section)
            for k, v in self.__md.getNodes().items():
                sectionModel.geometry.meshedConcrete.vertices[k] = (v.xn, v.yn)

            for k, v in self.__md.getTrianglesIds().items():
                sectionModel.geometry.meshedConcrete.triangles[k] = (v[0], v[1], v[2])

            # Rebars datas
            #
            for k, v in self.__md.getCircles().items():
                rebar = Rebar()
                rebar.idRebar = k
                rebar.idVertex = v.center.id
                rebar.radius = v.radius
                sectionModel.geometry.rebars[k] = rebar
            # ------------------------------------------------------------------
            #              M A T E R I A L S  F O R  S E C T I O N
            # ------------------------------------------------------------------
            # Steel material
            #
            if id_section in self.__matForRebars.keys():
                sectionModel.materials.rebars = self.__matForRebars[id_section]
            else:
                log(
                    "WRN",
                    f"Steel material not assigned to " f" section id *{id_section}*!!!",
                    self.__ll,
                )
            # Concrete material
            #
            if id_section in self.__matForConcrete.keys():
                sectionModel.materials.concrete = self.__matForConcrete[id_section]
            else:
                log(
                    "WRN",
                    f"Concrete material not assigned to "
                    f" section id *{id_section}*!!!",
                    self.__ll,
                )
            # ------------------------------------------------------------------
            #                           L O A D S
            # ------------------------------------------------------------------
            sectionModel.loads = self.__forces[id_section]
            modelInput.sections[id_section] = sectionModel
        return modelInput

    def __buildNodeId(self):
        self.__last_node_id += 1
        return self.__last_node_id

    def __buildSectionId(self):
        self.__last_section_id += 1
        return self.__last_section_id

    def __buildTriangleId(self):
        self.__last_triangle_id += 1
        return self.__last_triangle_id

    def __buildRebarId(self):
        self.__last_rebar_id += 1
        return self.__last_rebar_id

    def __buildMatId(self):
        self.__last_mat_id += 1
        return self.__last_mat_id

    @property
    def modeler(self) -> SectionModeler:
        return self.__md

    def rotate(self, degree: float, center: Point2d = Point2d(0.0, 0.0)):
        self.__md.rotate(degree, center)
        return

    def options(self) -> Options:
        return self.__options

    def posNode(self, idn: int) -> Point2d:
        return Point2d(self.__md.getNodeX(idn), self.__md.getNodeY(idn))

    def addRebarsGroup(
        self,
        points: List[Point2d],
        shape: ShapeCircle = None,
        diameter: float = None,
        id_nodes: List[int] = None,
        id_rebars: List[int] = None,
    ) -> List[int]:

        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return []

        if diameter is None and shape is None:
            log("ERR", "Not provided diameter or shape !!! Do nothing.", self.__ll)
            return []

        if diameter is not None and shape is None:
            d = diameter

        # shape is not None and diameter is None
        else:
            if shape.getType() != ShapesEnum.SHAPE_CIRC:
                log("ERR", "Shape subclass unknown or not implemented !!!", self.__ll)
                return []
            else:
                d = shape.getRadius() * 2

        log("INF", "Perform adding rebars", self.__ll)

        if id_nodes is not None and id_rebars is not None:
            if len(id_nodes) != len(points) or len(id_rebars) != len(points):
                log(
                    msg=f"Size of id nodes {len(id_nodes)} "
                    f"and id rebar {len(id_rebars)} must be equal to size "
                    f"of points {len(points)}!!!",
                    tp="ERR",
                    level=self.__ll,
                )
                return []

        barIds = []
        for i, c in enumerate(points):
            if id_nodes is None and id_rebars is None:
                idNode = self.__buildNodeId()
                idRebar = self.__buildRebarId()
            elif id_nodes is not None and id_rebars is not None:
                idNode = id_nodes[i]
                idRebar = id_rebars[i]
            else:
                log("ERR", "Wrong args !!!", self.__ll)
                return []

            if not self.__md.addNode(idNode, c.x, c.y):
                log(msg=f"Cannot add node {idNode} !!!", tp="ERR", level=self.__ll)
                return []

            if not self.__md.addCircle(idRebar, idNode, d / 2):
                log(msg=f"Cannot add rebar {idRebar} !!!", tp="ERR", level=self.__ll)
                return []

            barIds.append(idRebar)

        log(
            msg=f"Added group of {len(barIds)} rebars with node {barIds}.",
            tp="INF",
            level=self.__ll,
        )
        return barIds

    def addSectionModel(self, key: str, current: bool = True, ids: int = -1) -> int:

        if key in self.__sectionKeys.keys():
            log("ERR", "The key is yet present !!!", self.__ll)
            return -1

        if ids < 0:
            id_section = self.__buildSectionId()
        else:
            if ids in self.__sectionKeys.values():
                log("ERR", f"The id {ids} is yet present !!!", self.__ll)
                return -1
            else:
                id_section = ids

        self.__md.addSection(id_section, current)
        self.__sectionKeys[key] = id_section
        self.__results.sectionResults[id_section] = Results(sectionKey=key)

        if self.__forces is None:
            self.__forces = {id_section: {}}
        else:
            self.__forces[id_section] = {}

        return id_section

    def setCurrentModel(self, key: str = None, ids: int = None) -> bool:
        if ids is None and key is None:
            log(
                "ERR",
                "Key and ids are None type. Con't make current nothing !!!",
                self.__ll,
            )
            return False

        if ids is not None and key is None:
            if not self.__md.setCurrent(ids):
                log("ERR", f"The id {ids} isn't present !!!", self.__ll)
                return False
            return True

        if ids is None and key is not None:
            if key not in self.__sectionKeys.keys():
                log("ERR", "The key isn't present !!!", self.__ll)
                return False

        if self.__md.getCurrent() != self.__sectionKeys[key]:
            self.__md.setCurrent(self.__sectionKeys[key])

        return True

    def addNode(self, node: Node2d) -> int:
        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return False

        if node.idn >= 0:
            idn = node.idn
        else:
            idn = self.__buildNodeId()

        if self.__md.addNode(idn, node.x, node.y):
            return idn
        else:
            log(
                msg=f"Node with id={idn} cannot be added to model !!!",
                tp="ERR",
                level=self.__ll,
            )
            return -1

    def addTriangle(self, idn_1: int, idn_2: int, idn_3: int, idt: int = -1) -> int:

        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return False

        if idt >= 0:
            id_tria = idt
        else:
            id_tria = self.__buildTriangleId()

        if self.__md.addTriangle(id_tria, idn_1, idn_2, idn_3):
            if self.__trianglesIds is None:
                self.__trianglesIds = {self.__md.getCurrent(): [id_tria]}
            else:
                if self.__md.getCurrent() in self.__trianglesIds.keys():
                    self.__trianglesIds[self.__md.getCurrent()].append(id_tria)
                else:
                    self.__trianglesIds[self.__md.getCurrent()] = [id_tria]
            return id_tria
        else:
            log(
                msg=f"Triangle with id=[{id_tria}] and nodes "
                f"{idn_1} {idn_2} {idn_3} cannot be added to model !!!",
                tp="ERR",
                level=self.__ll,
            )
            return -1

    def addSolid(self, shape: Union[ShapeRect, ShapeCircle]) -> bool:
        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return False

        if shape.getType() == ShapesEnum.SHAPE_RECT:
            if self.__shapeDicts is None:
                self.__shapeDicts = {self.__md.getCurrent(): shape}
            else:
                self.__shapeDicts[self.__md.getCurrent()] = shape
            log("INF", "Perform build rectangular shape", self.__ll)
            idBL = self.__buildNodeId()
            pBL = shape.getShapePoint("BL")
            idBR = self.__buildNodeId()
            pBR = shape.getShapePoint("BR")
            idTR = self.__buildNodeId()
            pTR = shape.getShapePoint("TR")
            idTL = self.__buildNodeId()
            pTL = shape.getShapePoint("TL")
            if self.__namedVerticesDict is None:
                self.__namedVerticesDict = {
                    self.__md.getCurrent(): {
                        "BL": idBL,
                        "BR": idBR,
                        "TR": idTR,
                        "TL": idTL,
                    }
                }
            else:
                self.__namedVerticesDict[self.__md.getCurrent()] = {
                    "BL": idBL,
                    "BR": idBR,
                    "TR": idTR,
                    "TL": idTL,
                }

            self.__md.addNode(idBL, pBL.x, pBL.y)
            self.__md.addNode(idBR, pBR.x, pBR.y)
            self.__md.addNode(idTR, pTR.x, pTR.y)
            self.__md.addNode(idTL, pTL.x, pTL.y)
            idTria1 = self.__buildTriangleId()
            self.__md.addTriangle(idTria1, idBL, idBR, idTR)
            idTria2 = self.__buildTriangleId()
            self.__md.addTriangle(idTria2, idTR, idTL, idBL)

            if self.__trianglesIds is None:
                self.__trianglesIds = {self.__md.getCurrent(): [idTria1, idTria2]}
            else:
                self.__trianglesIds[self.__md.getCurrent()] = [idTria1, idTria2]

            return True

        log("ERR", "Shape subclass unknown or not implemented !!!", self.__ll)
        return False

    def getShapeVertex(self, name: str) -> Node2d:
        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return Node2d()
        isec = self.__md.getCurrent()

        if self.__shapeDicts is None:
            log("ERR", "Shape not assigned with addSolid() !!!", self.__ll)
            return Node2d()

        if self.__namedVerticesDict is None:
            log(
                "ERR",
                "Vertices are None. Luckly shape not assigned with addSolid() !!!",
                self.__ll,
            )
            return Node2d()
        else:
            if name not in self.__namedVerticesDict[isec].keys():
                log("ERR", f"Name {name} not in vertices keys !!!", self.__ll)
                return Node2d()
            else:
                idNode = self.__namedVerticesDict[isec][name]
                point = self.__shapeDicts[isec].getShapePoint(name)
                return Node2d(point.x, point.y, idNode)

    def addRebar(
        self,
        shape: Union[ShapeCircle, None] = None,
        center: Point2d = Point2d(),
        diameter: float = 0,
        idn: int = -1,
        idr: int = -1,
        use_vertex: bool = False
    ) -> int:

        if self.__md.getCurrent() == -1:
            log("ERR", "None model current !!!", self.__ll)
            return -1

        if not use_vertex:
            if shape is not None:
                if shape.getType() != ShapesEnum.SHAPE_CIRC:
                    log("ERR", "Shape subclass unknown or not implemented !!!", self.__ll)
                    return -1
                c = shape.getShapePoint("O")
                r = shape.getRadius()
            else:
                c = center
                r = diameter / 2

            if idn < 0:
                id_node = self.__buildNodeId()
            else:
                id_node = idn

            if not self.__md.addNode(id_node, c.x, c.y):
                log(
                    msg=f"Add single rebar with node {id_node} error !!!",
                    tp="ERR",
                    level=self.__ll,
                )
                return -1
            else:
                log(
                    msg=f"Add single rebar with node {id_node} with success",
                    tp="INF",
                    level=self.__ll,
                )
        else:
            id_node = idn
            r = diameter / 2

        if idr < 0:
            id_rebar = self.__buildRebarId()
        else:
            id_rebar = idr

        if not self.__md.addCircle(id_rebar, id_node, r):
            log(
                msg=f"Add single rebar with circle {id_rebar} error !!!",
                tp="ERR",
                level=self.__ll,
            )
            return -1
        else:
            log(
                msg=f"Add single rebar with circle {id_rebar} with success",
                tp="INF",
                level=self.__ll,
            )

        log(
            msg=f"Add single rebar with node {id_node} and rebar {id_rebar}",
            tp="INF",
            level=self.__ll,
        )

        return id_rebar

    def plot(self, pfx: str = "", onlyWorst: bool = True) -> None:
        outPath = super().outPath()
        log("INF", "Start plot ...", self.__ll)

        # Journal of command
        com = Command(type=CommandType.PLOT)
        com.args.append(CommandArg(name="pfx", value=pfx))
        com.args.append(CommandArg(name="onlyWorst", value=onlyWorst))
        self.__journalOfCommands.append(com)

        if outPath.exists():
            log("INF", f"... path {outPath} exist !!!", self.__ll)
        else:
            log("INF", f"... making path {outPath}", self.__ll)
            os.mkdir(outPath)
            if outPath.exists():
                log("INF", "... path maked successfully", self.__ll)
            else:
                log("ERR", "... something was wrong", self.__ll)
                return

        # Saving current model
        holdCurrentModel = self.__md.getCurrent()

        sp = SectionPlot()
        sp.logLevel = self.__ll
        sp.modeler = self.__md
        # ----------------------------------------------------------------------
        #                          PLOTTING GEOMETRIES
        # ----------------------------------------------------------------------
        # Iterate over model indexes
        #
        sp.setView(SectionPlotView_Enum.SP_VIEW_SOLID_NODES)
        for modelKey in self.__sectionKeys.keys():
            log("INF", f"... plot model with key = *{modelKey}*", self.__ll)
            self.__md.setCurrent(self.__sectionKeys[modelKey])

            # Plotting with tool
            #
            sp.plot()

            art_path = Path(modelKey)
            art_name = f"{pfx}geometry"

            (outPath / art_path).mkdir(exist_ok=True)
            sp.save(fname=outPath / art_path / art_name, dpi=300, transparent=False)
            res = self.__results.sectionResults[self.__sectionKeys[modelKey]]
            artifact = Artifact(
                tp=ArtifactType.SECTION_GEOMETRY,
                path=art_path.as_posix(),
                name=art_name,
                ext="png",
            )
            res.artifacts.append(artifact)

        if len(self.__results.sectionResults) > 0:
            log("INF", "... plot domains ...", self.__ll)
        else:
            log("INF", "... there isn't domains in results", self.__ll)

        for ids in self.__results.sectionResults:
            res = self.__results.sectionResults[ids]
            log("INF", f"... ... plot domain for section {res.sectionKey}", self.__ll)

            # ------------------------------------------------------------------
            #                       PLOTTING DOMAIN SLU
            # -------------------------------------------------------------------
            if res.domainSLU is None:
                log(
                    "WRN",
                    f"... ... ... domain doesn't calculated "
                    f"for section *{res.sectionKey}*",
                    self.__ll,
                )
            else:

                if onlyWorst:
                    keysToPlot = [res.domainSLU.loadIdCheck]
                else:
                    keysToPlot = res.domainSLU.loadsCheckLogs

                for keyLoad in keysToPlot:
                    loadres = res.domainSLU.loadsCheckLogs[keyLoad]
                    pointsCloud = PointCloud2d(loadres.domain.pointsCloud2d)
                    plotter = Geometry2DPlot(1, 1, figsize=(7, 7), sx=1e-6, sy=1e-6)
                    plotter.addPointCloud(pointsCloud)

                    pointArrayMy = [Point2d(loadres.Mxed, loadres.Myed)]
                    col = "red"
                    if loadres.check:
                        col = "green"
                    plotter.addPointArray(pointArrayMy, [col])
                    plotter.showItems(False, False, True)
                    plotter.setTitles(
                        [
                            f"Domain Mx-My [KNm] at N={loadres.Ned * 1e-3:.1f} [KN] - load id: {loadres.loadId}\n"
                            f"Nmin={loadres.Nmin * 1e-3:.1f} - "
                            f"Nmax={loadres.Nmax * 1e-3:.1f}",
                        ]
                    )
                    plotter.setXLabel(
                        [
                            "Mx [KNm]",
                        ]
                    )
                    plotter.setYLabel(
                        [
                            "My [KNm]",
                        ]
                    )
                    plotter.plot()
                    art_path = Path(res.sectionKey) / "domain"
                    art_name = f"{pfx}load_{loadres.loadId}_MxMy"

                    (outPath / art_path).mkdir(exist_ok=True)
                    plotter.save(fname=outPath / art_path / art_name, dpi=300)
                    artifact = Artifact(
                        tp=ArtifactType.DOMAIN_SLU,
                        path=art_path.as_posix(),
                        name=art_name,
                        ext="png",
                    )
                    if onlyWorst:
                        res.domainSLU.artifacts.append(artifact)
                        loadres.artifacts.append(artifact)
                    else:
                        loadres.artifacts.append(artifact)

            # ------------------------------------------------------------------
            #                    PLOTTING ELASTIC RESULTS
            # -------------------------------------------------------------------
            if res.elasticSolver is None:
                log(
                    "WRN",
                    f"... ... ... elastic solution don't calculated "
                    f"for section *{res.sectionKey}*",
                    self.__ll,
                )
            else:

                if onlyWorst and res.elasticCheck is not None:
                    keysToPlot = [res.elasticCheck.loadIdCheck]
                else:
                    keysToPlot = res.elasticSolver.elasticResults

                sp = SectionPlot()
                sp.logLevel = self.__ll
                sp.modeler = self.__md
                sp.setView(SectionPlotView_Enum.SP_VIEW_STRESSES)
                for keyLoad in keysToPlot:
                    loadres = res.elasticSolver.elasticResults[keyLoad]
                    sp.vertexStresses = loadres.sigmaci
                    sp.rebarStresses = loadres.sigmasi
                    self.__md.setCurrent(self.__sectionKeys[res.sectionKey])
                    sp.plot()

                    art_path = Path(res.sectionKey) / "stress"
                    art_name = f"{pfx}load_{keyLoad}_stress"

                    (outPath / art_path).mkdir(exist_ok=True)
                    sp.save(fname=outPath / art_path / art_name, dpi=300)
                    artifact = Artifact(
                        tp=ArtifactType.ELASTIC_STRESS,
                        path=art_path.as_posix(),
                        name=art_name,
                        ext="png",
                    )
                    if onlyWorst and res.elasticCheck is not None:
                        res.elasticCheck.artifacts.append(artifact)
                        loadres.artifacts.append(artifact)
                    else:
                        loadres.artifacts.append(artifact)

        log("INF", "Stop plot ...", self.__ll)

        # Restore held model
        self.__md.setCurrent(holdCurrentModel)

    def addConcreteLaw(
        self,
        mat: Concrete,
        shape: Literal["PARABOLA-RECTANGLE"] = "PARABOLA-RECTANGLE",
        idm: int = -1,
    ) -> int:

        if shape != "PARABOLA-RECTANGLE":
            log("ERR", "Law shape unknown", self.__ll)
            return -1

        if idm < 0:
            id_mat = self.__buildMatId()
        else:
            if idm in self.__concreteMaterial.keys():
                log("ERR", f"Material with id = {idm} exists !!!", self.__ll)
                return -1
            id_mat = idm

        mat.setId(id_mat)
        ret_val = self.__md.addLawParaboleRectangle(
            ids=mat.getId(),
            descr=mat.getMatDescr(),
            fcd=-mat.get_alphacc() * mat.get_fck() / mat.get_gammac(),
            ec2=-mat.get_ec2(),
            ecu=-mat.get_ecu(),
        )
        if not ret_val:
            log("ERR", f"Cannot add concrete law with id" f"{id_mat}", self.__ll)
            return -1

        self.__concreteMaterial[id_mat] = mat
        return id_mat

    def assignConcreteLawToCurrentModel(self, idm: int) -> bool:
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0014", "None model current")

        if idm not in self.__concreteMaterial.keys():
            raise Ex(
                "(templateRCGen)-0015",
                f"Concrete material with id *{idm}* doesn't exist.",
            )

        if self.__md.setTrianglesLawInCurrentModel(idm):
            self.__matForConcrete[self.__md.getCurrent()] = idm
            return True

        return False

    def addRebarLaw(
        self,
        mat: ConcreteSteel,
        shape: Literal["LINEAR-RECTANGLE"] = "LINEAR-RECTANGLE",
        idm: int = -1,
    ) -> int:

        if shape != "LINEAR-RECTANGLE":
            log("ERR", "Law shape unknown", self.__ll)
            return -1

        if idm < 0:
            id_mat = self.__buildMatId()
        else:
            if idm in self.__rebarsMaterial.keys():
                log("ERR", f"Material with id = {idm} exists !!!", self.__ll)
                return -1
            id_mat = idm

        mat.setId(id_mat)

        retVal = self.__md.addLawBilinear(
            ids=mat.getId(),
            descr=mat.getMatDescr(),
            fsd=mat.get_fsy() / mat.get_gammas(),
            Es=mat.get_Es(),
            esu=mat.get_esu(),
        )
        if not retVal:
            log("ERR", "Modeler error with addLawBilinear", self.__ll)
            return -1

        self.__rebarsMaterial[id_mat] = mat
        return id_mat

    def assignRebarLawToCurrentModel(self, idm: int) -> bool:
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0013", "None model current")

        if idm not in self.__rebarsMaterial.keys():
            raise Ex(
                "(templateRCGen)-0016",
                f"Rebars material with id *{idm}* doesn't exist.",
            )

        if self.__md.setCirclesLawInCurrentModel(idm):
            self.__matForRebars[self.__md.getCurrent()] = idm
            return True

        return False

    def calcConcreteArea(self) -> float:
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0009", "None model current")
        return self.__md.calcSolidArea()

    def calcRebarArea(self) -> float:
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0010", "None model current")

        return self.__md.calcPointArea()

    def assignForces(self, forces: Dict[int, Forces] = None) -> None:
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0012", "None model current")

        self.__forces[self.__md.getCurrent()] = forces

    def __checkBeforeRun(self):
        if self.__md.getCurrent() == -1:
            raise Ex("(templateRCGen)-0011", "None model current")

        if self.__md.getCurrent() not in self.__matForConcrete.keys():
            raise Ex(
                "(templateRCGen)-0015", "Concrete not assigned in current section !!!"
            )

        if self.__md.getCurrent() not in self.__matForRebars.keys():
            raise Ex(
                "(templateRCGen)-0016",
                "Rebar material not assigned in current section !!!",
            )

    def __run_check_domain_slu(self, res: Results) -> bool:
        res.domainSLU = CheckDomainSLUResults()
        res.domainSLU.logs += [
            log("INF", "... perform DOMAIN_SLU analysis ...", self.__ll)
        ]

        # I have to remove old results
        res.domainSLU = CheckDomainSLUResults()

        if self.__forces is None:
            res.domainSLU.logs += [log("ERR", "Using assignForces !!!", self.__ll)]
            return False

        # Check forces if null or yet assigned
        if self.__md.getCurrent() not in self.__forces.keys():
            res.domainSLU.logs += [
                log("ERR", "Forces in Domain SLU analysis are nulls !!!", self.__ll)
            ]
            return False

        forces = self.__forces[self.__md.getCurrent()]

        if len(forces) == 0:
            res.domainSLU.logs += [
                log("ERR", "Forces with lenght=0 !!! Stop.", self.__ll)
            ]
            return False

        res.domainSLU.logs += [
            log("INF", "Successfully assigned forces for analysis ...", self.__ll)
        ]

        # Check forces if null
        counterNull = 0
        counterSkipped = 0
        filteredForces: Dict[int, Forces] = {}

        for _k, f in forces.items():
            haveToSkip = False

            if f.isNull():
                counterNull += 1
                haveToSkip = False
            if f.limitState is not LimitState.ULTIMATE:
                counterSkipped += 1
                haveToSkip = True

            if not haveToSkip:
                filteredForces[f.id] = f

        res.domainSLU.globalCheck = CheckResults()
        if len(filteredForces) == 0:
            res.domainSLU.logs += [
                log(
                    "WRN",
                    f"All {len(forces)} forces are null or not SLU !!!",
                    self.__ll,
                )
            ]
            res.domainSLU.globalCheck.check = True
            res.domainSLU.globalCheck.safetyFactor = -1

        res.domainSLU.logs += [
            log(
                "INF",
                f"Nb. {counterNull}/{len(forces)} " f"forces are null !!!",
                self.__ll,
            )
        ]
        res.domainSLU.logs += [
            log(
                "INF",
                f"Nb. {counterSkipped}/{len(forces)} " f"forces are not ULTIMATE !!!",
                self.__ll,
            )
        ]

        checkVector: List[bool] = []
        safetyVector: List[float] = []
        loadsIdVector: List[int] = []
        loadsNOOK: List[int] = []
        loadsOK: List[int] = []

        for _k, force in filteredForces.items():
            check = DomainSLUCheck(loadId=force.id)
            checkLogs = DomainSLUCheckLog(
                Mxrd=0.0,
                Myrd=0.0,
                Nrd=0.0,
                loadId=force.id,
                Mxed=force.Mx,
                Myed=force.My,
                Ned=force.Fz,
                check=False,
                Nmin=0.0,
                Nmax=0.0,
            )
            check.interactionDomain.check = True
            check.interactionDomain.safetyFactor = "inf"
            # oldRef = force.ref
            # force.switchToNamedRef('DOMAIN')
            if force.isNull():
                checkLogs.logs += ["Force null"]
            else:
                degreeDivision = self.__options.domainSLU.degreeDivision
                rationDivision = self.__options.domainSLU.rationDivision
                rebuild = self.__options.domainSLU.rebuild
                if self.__md.buildDomain(degreeDivision, rationDivision, rebuild):
                    checkLogs.logs += ["Domain ok"]
                    res.domainSLU.logs += [
                        log(
                            "INF",
                            f"Domain formed successfully with degree {degreeDivision}, "
                            f"ratio {rationDivision}, rebuild {rebuild}",
                            self.__ll,
                        )
                    ]
                    minN = self.__md.domainBounding()[0]
                    maxN = self.__md.domainBounding()[1]
                    roX = abs(
                        self.__md.domainBounding()[3] - self.__md.domainBounding()[2]
                    )
                    roY = abs(
                        self.__md.domainBounding()[5] - self.__md.domainBounding()[4]
                    )

                    assert minN < 0.0
                    assert maxN > 0.0

                    checkLogs.Nmin = minN
                    checkLogs.Nmax = maxN

                    # Compute about safety factor for N
                    if force.Fz < 0.0:
                        safetyN = minN / force.Fz
                        checkLogs.Nrd = minN
                    elif force.Fz > 0.0:
                        safetyN = maxN / force.Fz
                        checkLogs.Nrd = maxN
                    else:
                        safetyN = "inf"
                        checkLogs.Nrd = None

                    if minN < force.Fz < maxN:
                        checkLogs.logs += ["N is contained in bounding"]
                        intersectionsN = self.__md.intersectAtN(force.Fz)
                        checkLogs.domain.pointsCloud2d = intersectionsN.model()
                        # print(f"Intersection plane at N={force.Fz}")
                        # print(f"for Mx={force.Mx} and My= {force.My} and ro=({roX},{roY})")
                        # intersectionsN = self.__md.intersectAtN(force.Fz, 1e-6, 1e-6)
                        if float(force.Mx) == 0.0 and float(force.My) == 0.0:
                            checkLogs.logs += [
                                "Moments Mx and My are nulls. Banal solution"
                            ]
                            check.interactionDomain.safetyFactor = safetyN
                            if check.interactionDomain.safetyFactor == "inf":
                                check.interactionDomain.check = True
                            else:
                                check.interactionDomain.check = (
                                    check.interactionDomain.safetyFactor > 1.0
                                )
                            checkLogs.Mxrd = None
                            checkLogs.Myrd = None
                            checkLogs.check = check.interactionDomain.check
                        else:
                            (
                                contained,
                                pintersect,
                                intfactor0,
                                pindex,
                            ) = intersectionsN.contains(
                                xp=force.Mx, yp=force.My, ro=(roX, roY)
                            )
                            assert intfactor0 >= 0
                            if safetyN == "inf":
                                check.interactionDomain.safetyFactor = intfactor0
                            else:
                                check.interactionDomain.safetyFactor = min(
                                    intfactor0, safetyN
                                )

                            check.interactionDomain.check = (
                                check.interactionDomain.safetyFactor > 1
                            )
                            checkLogs.Mxrd = intfactor0 * force.Mx
                            checkLogs.Myrd = intfactor0 * force.My
                            checkLogs.check = check.interactionDomain.check
                    else:
                        checkLogs.logs += ["N isn't contained in bounding"]

                else:
                    checkLogs.logs += ["Domain error !!!"]
                    res.domainSLU.logs += [log("ERR", "Domain error !!!", self.__ll)]

            res.domainSLU.loadsCheck[force.id] = check
            res.domainSLU.loadsCheckLogs[force.id] = checkLogs

            checkVector += [check.interactionDomain.check]

            if type(check.interactionDomain.safetyFactor) == float:
                safetyVector += [check.interactionDomain.safetyFactor]
                loadsIdVector += [force.id]
                if not check.interactionDomain.safetyFactor > 1:
                    loadsNOOK += [check.loadId]
                else:
                    loadsOK += [check.loadId]
            else:
                loadsOK += [check.loadId]

            # force.switchToRef(oldRef)

        if len(filteredForces) > 0:
            res.domainSLU.globalCheck.check = False not in checkVector
            if len(safetyVector) > 0:
                min_index = min(range(len(safetyVector)), key=lambda i: safetyVector[i])
                res.domainSLU.globalCheck.safetyFactor = safetyVector[min_index]
                res.domainSLU.loadIdCheck = loadsIdVector[min_index]
            # All forces are null then
            else:
                res.domainSLU.globalCheck.safetyFactor = "inf"

        return True

    def __run_check_elastic_sls(self, res: Results) -> bool:
        res.elasticCheck = ElasticSLSCheckResults()
        res.elasticCheck.logs += [
            log("INF", "... perform ELASTIC CHECK analysis ...", self.__ll)
        ]

        # Need run ELASTIC SOLVER before
        if res.elasticSolver is None:
            res.elasticCheck.logs += [
                log("ERR", "Need ELASTIC RUN before !!! Stop.", self.__ll)
            ]
            return False
        if len(res.elasticSolver.elasticResults) == 0:
            res.elasticCheck.logs += [
                log(
                    "ERR",
                    "Need ELASTIC RUN before with almost one" "load !!! Stop.",
                    self.__ll,
                )
            ]
            return False
        else:
            res.elasticCheck.logs += [
                log(
                    "INF",
                    f"Nb.{len(res.elasticSolver.elasticResults)}"
                    f" loads will be checked !!! Stop.",
                    self.__ll,
                )
            ]

        forces_on_section = self.__forces[self.__md.getCurrent()]

        concrete_mat_used = self.__concreteMaterial[
            self.__matForConcrete.get(self.__md.getCurrent())
        ]
        if concrete_mat_used is None:
            res.elasticCheck.logs += [
                log(
                    "ERR",
                    f"Concrete material don't assigned "
                    f"to section {self.__md.getCurrent()} !!! Stop.",
                    self.__ll,
                )
            ]
            return False
        rebar_mat_used = self.__rebarsMaterial[
            self.__matForRebars.get(self.__md.getCurrent())
        ]
        if rebar_mat_used is None:
            res.elasticCheck.logs += [
                log(
                    "ERR",
                    f"Rebars material don't assigned "
                    f"to section {self.__md.getCurrent()} !!! Stop.",
                    self.__ll,
                )
            ]
            return False

        checkVector: List[bool] = []
        safetyVector: List[float] = []
        loadsIdVector: List[int] = []

        for k in res.elasticSolver.elasticResults.keys():
            force = forces_on_section[k]

            # Load need to be CHARACTERISTIC or QUASI_PERMANENT
            ch = force.frequency == Frequency.CHARACTERISTIC
            qp = force.frequency == Frequency.QUASI_PERMANENT
            if ch or qp:
                res.elasticCheck.loadsCheck[k] = ElasticSLSCheck()
                res.elasticCheck.loadsCheckLogs[k] = ElasticSLSCheckLog()
                res.elasticCheck.loadsCheckLogs[k].frequency = force.frequency
                elastic_result = res.elasticSolver.elasticResults[k]
                check = False
                if ch:
                    sigmac_max = -concrete_mat_used.get_sigmac_max_c()
                    sigmas_max = rebar_mat_used.get_sigmas_max_c()
                    ratio_steel = elastic_result.sigmasMax / sigmas_max
                    ratio_concrete = elastic_result.sigmacMin / sigmac_max
                    res.elasticCheck.loadsCheckLogs[k].sigmacMinCH = sigmac_max
                    res.elasticCheck.loadsCheckLogs[k].sigmasMaxCH = sigmas_max
                    res.elasticCheck.loadsCheckLogs[
                        k
                    ].sigmacCH = elastic_result.sigmacMin
                    res.elasticCheck.loadsCheckLogs[
                        k
                    ].sigmasCH = elastic_result.sigmasMax
                    if max(ratio_steel, ratio_concrete) != 0:
                        safetyFactor = 1 / max(ratio_steel, ratio_concrete)
                    else:
                        safetyFactor = "inf"
                    if safetyFactor >= 1 or safetyFactor == "inf":
                        check = True
                else:
                    sigmac_max = -concrete_mat_used.get_sigmac_max_q()
                    res.elasticCheck.loadsCheckLogs[k].sigmacMinQP = sigmac_max
                    ratio_concrete = elastic_result.sigmacMin / sigmac_max
                    res.elasticCheck.loadsCheckLogs[
                        k
                    ].sigmacQP = elastic_result.sigmacMin
                    if ratio_concrete != 0:
                        safetyFactor = 1 / ratio_concrete
                    else:
                        safetyFactor = "inf"
                    if safetyFactor >= 1 or safetyFactor == "inf":
                        check = True

                res.elasticCheck.loadsCheck[k].loadId = force.id
                res.elasticCheck.loadsCheck[k].elasticCheck.safetyFactor = safetyFactor
                res.elasticCheck.loadsCheck[k].elasticCheck.check = check
                checkVector += [check]
                if type(safetyFactor) is float:
                    safetyVector += [safetyFactor]
                    loadsIdVector += [force.id]

            else:
                res.elasticCheck.logs += [
                    log(
                        "INF",
                        f"Load {k} skipped because not "
                        f"CHARACTERISTIC or QUASI_PERMANENT !!! Skipped.",
                        self.__ll,
                    )
                ]

        if len(res.elasticCheck.loadsCheck) == 0:
            res.elasticCheck.logs += [
                log("ERR", "Forces all skipped !!! Stop.", self.__ll)
            ]
            return False

        min_index = min(range(len(safetyVector)), key=lambda i: safetyVector[i])
        res.elasticCheck.globalCheck = CheckResults()
        res.elasticCheck.globalCheck.safetyFactor = safetyVector[min_index]
        res.elasticCheck.loadIdCheck = loadsIdVector[min_index]

        res.elasticCheck.logs += [
            log("INF", "... ELASTIC CHECK end with success.", self.__ll)
        ]
        return True

    def __run_elastic_solver(self, res: Results) -> bool:
        res.elasticSolver = ElasticSolverResults()
        res.elasticSolver.logs += [
            log("INF", "... perform ELASTIC analysis ...", self.__ll)
        ]

        # Forces assigned to current section
        #
        forces = self.__forces[self.__md.getCurrent()]

        if len(forces) == 0:
            res.elasticSolver.logs += [
                log("ERR", "Forces with lenght=0 !!! Stop.", self.__ll)
            ]
            return False

        # Check forces if null
        counterNull = 0
        counterSkipped = 0
        filteredForces: Dict[int, Forces] = {}

        for _k, f in forces.items():
            haveToSkip = False

            if f.isNull():
                counterNull += 1
                haveToSkip = True
            if f.limitState is not LimitState.SERVICEABILITY:
                counterSkipped += 1
                haveToSkip = True

            if not haveToSkip:
                filteredForces[f.id] = f

        res.elasticSolver.logs += [
            log(
                "INF",
                f"Nb. {counterNull}/{len(forces)} " f"forces are null !!!",
                self.__ll,
            )
        ]
        res.elasticSolver.logs += [
            log(
                "INF",
                f"Nb. {counterSkipped}/{len(forces)} "
                f"forces are not SERVICEABILITY !!!",
                self.__ll,
            )
        ]

        for _k, force in filteredForces.items():
            # oldRef = force.ref
            # force.switchToNamedRef('ELASTIC')
            if not self.__md.elSolve(force=force, uncracked=False):
                res.domainSLU.logs += [
                    log(
                        "WRN",
                        f"... ELASTIC solver error with force id = {force.id}!!! Stop.",
                        self.__ll,
                    )
                ]
                return False
            resElastic = ElasticSolverResult()
            # force.switchToRef(oldRef)
            resElastic.Ned = force.Fz
            resElastic.Myed = force.My
            resElastic.Mxed = force.Mx
            (
                resElastic.sigmacMin,
                resElastic.sigmacMax,
            ) = self.__md.elExtremeStressConcrete()
            (
                resElastic.sigmasMin,
                resElastic.sigmasMax,
            ) = self.__md.elExtremeStressSteel()
            resElastic.state = self.__md.elSectionState()
            resElastic.tensionPlane = self.__md.elTensionPlane()

            # Results for solid nodes
            #
            solidNodes = self.__md.getSolidNodes()
            for v in solidNodes.keys():
                resElastic.sigmaci[v] = self.__md.elStressConcreteNodeId(v)

            # Results for rebars
            #
            rebars = self.__md.getCircles()
            for r in rebars.keys():
                resElastic.sigmasi[r] = self.__md.elStressSteelCircleId(r)

            res.elasticSolver.elasticResults[force.id] = resElastic

        res.elasticSolver.logs += [
            log("INF", "... ELASTIC analysis terminated successfully.", self.__ll)
        ]
        return True

    def __run_bounding_slu(self, res: Results) -> bool:
        # I have to remove old results
        res.boundingSLU = BoundingSLUResults()

        res.boundingSLU.logs += [
            log("INF", "... perform BOUNDING_SLU analysis ...", self.__ll)
        ]
        degreeDivision = self.__options.boundingSLU.degreeDivision
        rationDivision = self.__options.boundingSLU.rationDivision

        retVal = self.__md.buildDomain(
            degreeDivision, rationDivision, self.__options.boundingSLU.rebuild
        )
        if not retVal:
            res.boundingSLU.logs += [
                log("ERR", "... domain built with error ... return", self.__ll)
            ]
            return False

        res.boundingSLU.logs += [
            log("INF", "... domain built successfully ...", self.__ll)
        ]
        bbox = self.__md.domainBounding()

        res.boundingSLU.minForces = Forces(Fz=bbox[0], Mx=bbox[2], My=bbox[4])
        res.boundingSLU.maxForces = Forces(Fz=bbox[1], Mx=bbox[3], My=bbox[5])
        return True

    def run(self, opt: Union[Enum, None] = None, **kwargs) -> bool:

        self.__checkBeforeRun()

        res = self.__results.sectionResults[self.__md.getCurrent()]
        super()._setModelOutput(self.__results)

        res.logs += [
            log(
                "INF",
                f"Start analysis ... for section with id *{self.__md.getCurrent()}*",
                self.__ll,
            )
        ]

        com = Command(type=CommandType.RUN)
        if opt == Analysis.CHECK_DOMAIN_SLU:
            if not self.__run_check_domain_slu(res):
                return False
            arg = CommandArg(name="opt", value=Analysis.CHECK_DOMAIN_SLU)
            com.args.append(arg)

        if opt == Analysis.BOUNDING_SLU:
            if not self.__run_bounding_slu(res):
                return False
            arg = CommandArg(name="opt", value=Analysis.BOUNDING_SLU)
            com.args.append(arg)

        if opt == Analysis.CHECK_ELASTIC:
            if not self.__run_check_elastic_sls(res):
                return False
            arg = CommandArg(name="opt", value=Analysis.CHECK_ELASTIC)
            com.args.append(arg)

        if opt == Analysis.ELASTIC_SOLVER:
            if not self.__run_elastic_solver(res):
                return False
            arg = CommandArg(name="opt", value=Analysis.ELASTIC_SOLVER)
            com.args.append(arg)

        self.__journalOfCommands.append(com)

        res.logs += [log("INF", "End analysis.", self.__ll)]
        return True

    def buildReport(self, opt: Union[Enum, None] = None, **kwargs) -> bool:
        pass

class ReportElastic(FragmentsBuilder):
    def __init__(self, oDataSolver: ElasticSolverResults, oDataChecker: ElasticSLSCheckResults, jobPath: Path):
        super().__init__()
        self.__oDataSolver = oDataSolver
        self.__oDataChecker = oDataChecker
        self._setFragmentOptions({"job_path": jobPath})

    def buildFragment(self) -> Fragment:
        f = Fragment(getTemplatesPath())
        place_holder = dict()
        place_holder["check_SLE_MN"] = []
        artifacts = []
        for key_load in self.__oDataChecker.loadsCheckLogs.keys():
            check_logs = self.__oDataChecker.loadsCheckLogs[key_load]
            solver_res = self.__oDataSolver.elasticResults[key_load]
            check = self.__oDataChecker.loadsCheck[key_load]
            row_check = dict()
            row_check["id"] = key_load
            row_check["Ned"] = f"{solver_res.Ned / 1e3:.1f}"
            row_check["Mxed"] = f"{solver_res.Mxed / 1e6:.1f}"
            row_check["Myed"] = f"{solver_res.Myed / 1e6:.1f}"
            row_check["sigmas"] = f"{solver_res.sigmasMax:.1f}"
            row_check["sigmac"] = f"{solver_res.sigmacMin:.1f}"
            if check_logs.frequency == Frequency.CHARACTERISTIC:
                row_check["sigmaxs"] = f"{check_logs.sigmasMaxCH:.1f}"
                row_check["sigmaxc"] = f"{check_logs.sigmacMinCH:.1f}"
            elif check_logs.frequency == Frequency.QUASI_PERMANENT:
                row_check["sigmaxs"] = ""
                row_check["sigmaxc"] = f"{check_logs.sigmacMinQP:.1f}"
            else:
                row_check["sigmaxs"] = ""
                row_check["sigmaxc"] = ""
            if check.elasticCheck.check:
                row_check["check"] = "OK"
            else:
                row_check["check"] = "NOOK"
            row_check["FS"] = f"{check.elasticCheck.safetyFactor:.2f}"
            place_holder["check_SLE_MN"].append(row_check)

            for artifact in solver_res.artifacts:
                if artifact.tp == ArtifactType.ELASTIC_STRESS:
                    geometryFilePath_encoded_for_latex = str(str(self.getFragmentOptions()["job_path"] / artifact.file_name())).replace('\\', '/')
                    artifacts.append(
                        {
                            "domainUrl": geometryFilePath_encoded_for_latex,
                            "idLoad": key_load,
                            "newLine": False
                        }
                    )

        # TABLE for figures
        #
        maxCols = 2
        maxRows = 3
        max_elem_for_figure = maxCols * maxRows
        nb_elem = len(artifacts)
        nb_figs = math.ceil(nb_elem / max_elem_for_figure)

        figures_matrix = []
        elem = []
        fig_index = 0

        if nb_elem == 1:
            figure_width = 0.90
        else:
            figure_width = 0.41

        for a in artifacts:
            elem.append(a)
            if math.ceil(len(elem) / maxCols) == math.floor(len(elem) / maxCols):
                a["newLine"] = True
            if len(elem) == max_elem_for_figure:
                fig_index += 1
                figures_matrix.append({"domainUrls": elem, "figIndex": fig_index})
                elem = []
        if 0 < len(elem) < max_elem_for_figure:
            fig_index += 1
            figures_matrix.append({"domainUrls": elem, "figIndex": fig_index})

        place_holder["figures"] = figures_matrix
        place_holder["nbFigs"] = nb_figs
        place_holder["figuresWidth"] = figure_width

        f.add_template("template-ita-rc-gen-sle-nm.tex", place_holder)
        return f

class ReportBuilder(FragmentsBuilder):
    def __init__(self, iData: RCGenSectionsInput, oData: RCGenSectionsOutput, jobPath: Path):
        super().__init__()
        self.__iData = iData
        self.__oData = oData
        self.__jobPath = jobPath

    def buildFragment(self) -> Fragment:
        f = Fragment(getTemplatesPath())

        for key_section, section in self.__iData.sections.items():
            title_str = f"{key_section}"
            f.add(line=r"\section{Sezione n." + title_str + r"}")
            oData = self.__oData.sectionResults[key_section]
            #
            # Build GEOMETRY
            #
            for a in oData.artifacts:
                if a.tp == ArtifactType.SECTION_GEOMETRY:
                    section.setFragmentOptions({
                        "geometry_file_path": self.__jobPath / a.file_name()
                    })
            f.add(lines=section.buildFragment().frags())
            f.add_line(r"\subsection{Caratteristiche meccaniche dei materiali}")
            #
            # Build MATERIALS paragraph
            #
            # concrete
            #
            concrete_key = section.materials.concrete
            concrete_model = self.__iData.matCatalogue.concrete[concrete_key]
            concrete_fb = ConcreteFB(getTemplatesPath(), concrete_model.toMaterial())
            concrete_fb.setFragmentOptions(
                {
                    "section_title": "Calcestruzzo",
                    "section_level": EnumFBSection.SEC_SUBSUBSECTION,
                }
            )
            f.add(lines=concrete_fb.buildFragment().frags())
            #
            # steel
            #
            steel_key = section.materials.rebars
            steel_model = self.__iData.matCatalogue.steel[steel_key]
            steel_fb = SteelConcreteFB(getTemplatesPath(), steel_model.toMaterial())
            steel_fb.setFragmentOptions(
                {
                    "section_title": "Acciaio in barre",
                    "section_level": EnumFBSection.SEC_SUBSUBSECTION,
                }
            )
            f.add(lines=steel_fb.buildFragment().frags())
            #
            # Build FORCES paragraph
            #
            forces = []
            for v in section.loads.items():
                forces.append(v[1])
            forces_fb = ForcesOnSectionListFB(getTemplatesPath(), forces)
            forces_fb.setFragmentOptions(
                {
                    "section_title": "Carichi e combinazioni",
                    "section_level": EnumFBSection.SEC_SUBSECTION,
                }
            )
            f.add(lines=forces_fb.buildFragment().frags())
            #
            # Build SLU DOMAIN paragraph
            #
            if self.__oData.sectionResults[key_section].domainSLU is not None:
                self.__oData.sectionResults[key_section].domainSLU.setFragmentOptions({"job_path": self.__jobPath})
                f.add_line(r"\subsection{Verifiche agli SLU}")
                f.add(lines=self.__oData.sectionResults[key_section].domainSLU.buildFragment().frags())
                f.add_line(r"\newpage")

            #
            # Build SLE NM check
            #
            solver = self.__oData.sectionResults[key_section].elasticSolver
            checker = self.__oData.sectionResults[key_section].elasticCheck
            if solver is not None and checker is not None:
                elasticBuilder = ReportElastic(solver, checker, jobPath=self.__jobPath)
                f.add_line(r"\subsection{Verifiche agli SLE}")
                f.add(lines=elasticBuilder.buildFragment().frags())
                f.add_line(r"\newpage")


        #
        # Build CODE Paragraph
        #
        codesFragment = CodesFB(getTemplatesPath())
        for k in self.__iData.matCatalogue.concrete:
            mat_str = self.__iData.matCatalogue.concrete[k].code.value
            codesFragment.appendUniqueCode(code_str=mat_str)

        for k in self.__iData.matCatalogue.steel:
            mat_str = self.__iData.matCatalogue.steel[k].code.value
            codesFragment.appendUniqueCode(code_str=mat_str)

        f.add(lines=codesFragment.buildFragment().frags())
        return f
